import numpy.testing as npt
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

import ixmp
from ixmp.testing import assert_logs, make_dantzig, models


# Fixtures
@pytest.fixture(scope="class")
def scen(mp):
    """The default version of the Dantzig on the mp."""
    yield ixmp.Scenario(mp, **models["dantzig"])


@pytest.fixture(scope="function")
def scen_empty(request, test_mp):
    """An empty Scenario with a temporary name on the test_mp."""
    yield ixmp.Scenario(
        test_mp,
        model=request.node.nodeid.replace("/", " "),
        scenario="test",
        version="new",
    )


@pytest.fixture(scope="function")
def test_dict():
    return {
        "test_string": "test12345",
        "test_number": 123.456,
        "test_number_negative": -123.456,
        "test_int": 12345,
        "test_bool": True,
        "test_bool_false": False,
    }


class TestScenario:
    """Tests of :class:`ixmp.Scenario`."""

    # Initialize Scenario
    def test_init(self, test_mp, scen_empty):
        # Empty scenario has version == 0
        assert scen_empty.version == 0

        # A scenario with scheme='MESSAGE' can only be created with a subclass
        class Scenario(ixmp.Scenario):
            pass

        scen2 = Scenario(
            test_mp, model="foo", scenario="bar", scheme="MESSAGE", version="new"
        )

        # JDBCBackend complains unless these items are added
        scen2.add_set("technology", "t")
        scen2.add_set("year", "2000")
        scen2.commit("")
        del scen2

        # Loading this "MESSAGE-scheme" scenario with ixmp.Scenario raises an
        # exception
        with pytest.raises(RuntimeError):
            ixmp.Scenario(test_mp, model="foo", scenario="bar")

        # …but loading with a subclass of ixmp.Scenario is fine
        Scenario(test_mp, model="foo", scenario="bar")

    def test_default_version(self, mp):
        scen = ixmp.Scenario(mp, **models["dantzig"])
        assert scen.version == 2

    def test_from_url(self, mp, caplog):
        url = f"ixmp://{mp.name}/Douglas Adams/Hitchhiker"

        # Default version is loaded
        scen, mp = ixmp.Scenario.from_url(url)
        assert scen.version == 1

        # Giving an invalid version with errors='raise' raises an exception
        expected = (
            "There exists no Scenario 'Douglas Adams|Hitchhiker' "
            "(version: 10000)  in the database!"
        )
        with pytest.raises(Exception, match=expected):
            scen, mp = ixmp.Scenario.from_url(url + "#10000", errors="raise")

        # Giving an invalid scenario with errors='warn' raises an exception
        msg = (
            "ValueError: scenario='Hitchhikerfoo'\n"
            f"when loading Scenario from url: {repr(url + 'foo')}"
        )
        with assert_logs(caplog, msg):
            scen, mp = ixmp.Scenario.from_url(url + "foo")
        assert scen is None and isinstance(mp, ixmp.Platform)

    # Clone Scenario
    def test_clone(self, mp):
        scen = ixmp.Scenario(mp, **models["dantzig"], version=1)
        scen.remove_solution()
        scen.check_out()
        scen.init_set("h")
        scen.add_set("h", "test")
        scen.commit("adding an index set 'h', with element 'test'")

        scen2 = scen.clone(keep_solution=False)
        # Cloned scenario contains added set
        obs = scen2.set("h")
        npt.assert_array_equal(obs, ["test"])

    def test_clone_edit(self, scen):
        scen2 = scen.clone(keep_solution=False)
        scen2.check_out()
        scen2.change_scalar("f", 95.0, "USD/km")
        scen2.commit("change transport cost")

        # Original scenario that was clone is untouched
        assert scen.scalar("f") == {"unit": "USD/km", "value": 90}

        # Value is changed only in the clone
        assert scen2.scalar("f") == {"unit": "USD/km", "value": 95}

    # Initialize items
    def test_init_set(self, scen):
        """Test ixmp.Scenario.init_set()."""

        # Add set on a locked scenario
        with pytest.raises(
            RuntimeError,
            match="This Scenario cannot be edited" ", do a checkout first!",
        ):
            scen.init_set("foo")

        scen = scen.clone(keep_solution=False)
        scen.check_out()
        scen.init_set("foo")

        # Initialize an already-existing set
        with pytest.raises(ValueError, match="'foo' already exists"):
            scen.init_set("foo")

    def test_init_par(self, scen):
        scen = scen.clone(keep_solution=False)
        scen.check_out()

        # Parameter can be initialized with a tuple (not list) of idx_sets
        scen.init_par("foo", idx_sets=("i", "j"))

        # Return type of idx_sets is still list
        assert scen.idx_sets("foo") == ["i", "j"]

    def test_init_scalar(self, scen):
        scen2 = scen.clone(keep_solution=False)
        scen2.check_out()
        scen2.init_scalar("g", 90.0, "USD/km")
        scen2.commit("adding a scalar 'g'")

    # Existence checks
    def test_has_par(self, scen):
        assert scen.has_par("f")
        assert not scen.has_par("m")

    def test_has_set(self, scen):
        assert scen.has_set("i")
        assert not scen.has_set("k")

    def test_has_var(self, scen):
        assert scen.has_var("x")
        assert not scen.has_var("y")

    def test_scalar(self, scen):
        assert scen.scalar("f") == {"unit": "USD/km", "value": 90}

    # Store data
    @pytest.mark.parametrize(
        "args, kwargs",
        (
            # Scalar values/units are broadcast across multiple keys
            (("b", ["new-york", "chicago"]), dict(value=100, unit="cases")),
            # Empty DataFrame can be added without error
            (("b", pd.DataFrame(columns=["i", "j", "value", "unit"])), dict()),
            # Exceptions
            pytest.param(
                ("b", pd.DataFrame(columns=["i", "j", "value", "unit"])),
                dict(value=1.0),
                marks=pytest.mark.xfail(
                    raises=ValueError,
                    reason="both key_or_data.value and value supplied",
                ),
            ),
            pytest.param(
                ("b", pd.DataFrame(columns=["i", "j", "unit"])),
                dict(),
                marks=pytest.mark.xfail(
                    raises=ValueError, reason="no parameter values"
                ),
            ),
        ),
    )
    def test_add_par(self, scen, args, kwargs):
        scen = scen.clone(keep_solution=False)
        scen.check_out()
        scen.add_par(*args, **kwargs)

    def test_add_par2(self, scen):
        scen = scen.clone(keep_solution=False)
        scen.check_out()
        # Create a parameter with duplicate indices
        scen.init_par("foo", idx_sets=["i", "i", "j"], idx_names=["i0", "i1", "j"])
        scen.add_par("foo", pd.DataFrame(columns=["i0", "i1", "j"]), value=1.0)

    # Retrieve data
    def test_idx(self, scen):
        assert scen.idx_sets("d") == ["i", "j"]
        assert scen.idx_names("d") == ["i", "j"]

    def test_par(self, scen):
        # Parameter data can be retrieved with filters
        df = scen.par("d", filters={"i": ["seattle"]})
        # Units are as expected
        assert df.loc[0, "unit"] == "km"

    def test_items(self, scen):
        # Without filters
        iterator = scen.items()

        # next() can be called → an iterator was returned
        next(iterator)

        # Iterator returns the expected parameter names
        exp = ["a", "b", "d", "f"]
        for i, (name, data) in enumerate(scen.items()):
            # Name is correct in the expected order
            assert exp[i] == name
            # Data is one of the return types of .par()
            assert isinstance(data, (pd.DataFrame, dict))

        # Total number of items was correct
        assert i == 3

        # With filters
        iterator = scen.items(filters=dict(i=["seattle"]))
        exp = [("a", 1), ("d", 3)]
        for i, (name, data) in enumerate(iterator):
            # Name is correct in the expected order
            assert exp[i][0] == name
            # Number of (filtered) rows is as expected
            assert exp[i][1] == len(data)

        assert i == 1

    def test_var(self, scen):
        df = scen.var("x", filters={"i": ["seattle"]})

        # Labels along the 'j' dimension
        npt.assert_array_equal(df["j"], ["new-york", "chicago", "topeka"])
        # Levels
        npt.assert_array_almost_equal(df["lvl"], [50, 300, 0])
        # Marginals
        npt.assert_array_almost_equal(df["mrg"], [0, 0, 0.036])

    def test_load_scenario_data(self, mp):
        """load_scenario_data() caches all data."""
        scen = ixmp.Scenario(mp, **models["dantzig"])
        scen.load_scenario_data()

        cache_key = scen.platform._backend._cache_key(scen, "par", "d")

        # Item exists in cache
        assert cache_key in scen.platform._backend._cache

        # Cache has not been used
        hits_before = scen.platform._backend._cache_hit.get(cache_key, 0)
        assert hits_before == 0

        # Retrieving the expected value
        assert "km" == scen.par("d", filters={"i": ["seattle"]}).loc[0, "unit"]

        # Cache was used to return the value
        hits_after = scen.platform._backend._cache_hit[cache_key]
        assert hits_after == hits_before + 1

    def test_load_scenario_data_clear_cache(self, mp):
        # this fails on commit: 4376f54
        scen = ixmp.Scenario(mp, **models["dantzig"])
        scen.load_scenario_data()
        scen.platform._backend.cache_invalidate(scen, "par", "d")

    # I/O
    def test_excel_io(self, scen, scen_empty, tmp_path, caplog):
        tmp_path /= "output.xlsx"

        # FIXME remove_solution, check_out, commit, solve, commit should not
        #       be needed to make this small data addition.
        scen.remove_solution()
        scen.check_out()

        # A 1-D set indexed by another set
        scen.init_set("foo", "j")
        scen.add_set("foo", [["new-york"], ["topeka"]])
        # A scalar parameter with unusual units
        scen.platform.add_unit("pounds")
        scen.init_scalar("bar", 100, "pounds")
        # A parameter with no values
        scen.init_par("baz_1", ["i", "j"])
        # A parameter with ambiguous index name
        scen.init_par("baz_2", ["i"], ["i_dim"])
        scen.add_par("baz_2", dict(value=[1.1], i_dim=["seattle"]))
        # A 2-D set with ambiguous index names
        scen.init_set("baz_3", ["i", "i"], ["i", "i_also"])
        scen.add_set("baz_3", [["seattle", "seattle"]])
        # A set with no elements
        scen.init_set("foo_2", ["j"])

        scen.commit("")
        scen.solve()

        # Solved Scenario can be written to file
        scen.to_excel(tmp_path, items=ixmp.ItemType.MODEL)

        # With init_items=False, can't be read into an empty Scenario.
        # Exception raised is the first index set, alphabetically
        with pytest.raises(ValueError, match="no set 'i'; " "try init_items=True"):
            scen_empty.read_excel(tmp_path)

        # File can be read with init_items=True
        scen_empty.read_excel(tmp_path, init_items=True, commit_steps=True)

        # Contents of the Scenarios are the same, except for unreadable items
        assert set(scen_empty.par_list()) | {"baz_1", "baz_2"} == set(scen.par_list())
        assert set(scen_empty.set_list()) | {"baz_3"} == set(scen.set_list())
        assert_frame_equal(scen_empty.set("foo"), scen.set("foo"))
        # NB could make a more exact comparison of the Scenarios

        # Pre-initialize skipped items 'baz_2' and 'baz_3'
        scen_empty.check_out()
        scen_empty.init_par("baz_2", ["i"], ["i_dim"])
        scen_empty.init_set("baz_3", ["i", "i"], ["i", "i_also"])

        # Data can be read into an existing Scenario without init_items or
        # commit_steps arguments
        scen_empty.read_excel(tmp_path)

        # Re-initialize an item with different index names
        scen_empty.check_out()
        scen_empty.remove_par("d")
        scen_empty.init_par("d", idx_sets=["i", "j"], idx_names=["I", "J"])

        # Reading now logs an error about conflicting dims
        with assert_logs(caplog, "Existing par 'd' has index names(s)"):
            scen_empty.read_excel(tmp_path, init_items=True)

        # A new, empty Platform (different from the one under scen -> mp ->
        # test_mp) that lacks all units
        mp = ixmp.Platform(
            backend="jdbc", driver="hsqldb", url="jdbc:hsqldb:mem:excel_io"
        )
        # A Scenario without the 'dantzig' scheme -> no contents at all
        s = ixmp.Scenario(
            mp, model="foo", scenario="bar", scheme="empty", version="new"
        )

        # Fails with add_units=False
        with pytest.raises(
            ValueError, match="The unit 'pounds' does not exist" " in the database!"
        ):
            s.read_excel(tmp_path, init_items=True)

        # Succeeds with add_units=True
        s.read_excel(tmp_path, add_units=True, init_items=True)

    # Combined tests
    def test_meta(self, mp, test_dict):
        scen = ixmp.Scenario(mp, **models["dantzig"], version=1)
        for k, v in test_dict.items():
            scen.set_meta(k, v)

        # test all
        obs_dict = scen.get_meta()
        for k, exp in test_dict.items():
            obs = obs_dict[k]
            assert obs == exp

        # test name
        obs = scen.get_meta("test_string")
        exp = test_dict["test_string"]
        assert obs == exp

        scen.remove_meta(["test_int", "test_bool"])
        obs = scen.get_meta()
        assert len(obs) == 4
        assert set(obs.keys()) == {
            "test_string",
            "test_number",
            "test_number_negative",
            "test_bool_false",
        }

        # Setting with a type other than int, float, bool, str raises TypeError
        with pytest.raises(ValueError, match="Cannot use value"):
            scen.set_meta("test_string", complex(1, 1))

    def test_meta_bulk(self, mp, test_dict):
        scen = ixmp.Scenario(mp, **models["dantzig"], version=1)
        scen.set_meta(test_dict)

        # test all
        obs_dict = scen.get_meta()
        for k, exp in test_dict.items():
            obs = obs_dict[k]
            assert obs == exp

        # check updating metadata (replace and append)
        scen.set_meta({"test_int": 1234567, "new_attr": "new_attr"})
        assert scen.get_meta("test_int") == 1234567
        assert scen.get_meta("new_attr") == "new_attr"


def test_range(scen_empty):
    scen = scen_empty

    scen.init_set("ii")
    ii = range(1, 20, 2)

    # range instance is automatically converted to list of str in add_set
    scen.add_set("ii", ii)

    scen.init_par("new_par", idx_sets="ii")

    # range instance is a valid key argument to add_par
    scen.add_par("new_par", ii, [1.2] * len(ii))


def test_gh_210(scen_empty):
    scen = scen_empty
    i = ["i0", "i1", "i2"]

    scen.init_set("i")
    scen.add_set("i", i)
    scen.init_par("foo", idx_sets="i")

    columns = ["i", "value"]
    foo_data = pd.DataFrame(zip(i, [10, 20, 30]), columns=columns)

    # foo_data is not modified by add_par()
    scen.add_par("foo", foo_data)
    assert all(foo_data.columns == columns)


def test_set(scen_empty):
    """Test ixmp.Scenario.add_set(), .set(), and .remove_set()."""
    scen = scen_empty

    # Add element to a non-existent set
    with pytest.raises(KeyError, match=repr("foo")):
        scen.add_set("foo", "bar")

    # Initialize a 0-D set
    scen.init_set("i")

    # Add elements to a 0-D set
    scen.add_set("i", "i1")  # Name only
    scen.add_set("i", "i2", "i2 comment")  # Name and comment
    scen.add_set("i", ["i3"])  # List of names, length 1
    scen.add_set("i", ["i4", "i5"])  # List of names, length >1
    scen.add_set("i", range(0, 3))  # Generator (range object)
    # Lists of names and comments, length 1
    scen.add_set("i", ["i6"], ["i6 comment"])
    # Lists of names and comments, length >1
    scen.add_set("i", ["i7", "i8"], ["i7 comment", "i8 comment"])

    # Incorrect usage

    # Lists of different length
    with pytest.raises(ValueError, match="Comment 'extra' without matching key"):
        scen.add_set("i", ["i9"], ["i9 comment", "extra"])
    with pytest.raises(ValueError, match="Key 'extra' without matching comment"):
        scen.add_set("i", ["i9", "extra"], ["i9 comment"])

    # Add elements to a 1D set
    scen.init_set("foo", "i", "dim_i")
    scen.add_set("foo", ["i1"])  # Single key
    scen.add_set("foo", ["i2"], "i2 in foo")  # Single key and comment
    scen.add_set("foo", "i3")  # Bare name automatically wrapped
    # Lists of names and comments, length 1
    scen.add_set("foo", ["i6"], ["i6 comment"])
    # Lists of names and comments, length >1
    scen.add_set("foo", [["i7"], ["i8"]], ["i7 comment", "i8 comment"])
    # Dict
    scen.add_set("foo", dict(dim_i=["i7", "i8"]))

    # Incorrect usage
    # Improperly wrapped keys
    with pytest.raises(
        ValueError,
        match=r"2-D key \['i4', 'i5'\] invalid for " r"1-D set foo\['dim_i'\]",
    ):
        scen.add_set("foo", ["i4", "i5"])
    with pytest.raises(ValueError):
        scen.add_set("foo", range(0, 3))
    # Lists of different length
    with pytest.raises(ValueError, match="Comment 'extra' without matching key"):
        scen.add_set("i", ["i9"], ["i9 comment", "extra"])
    with pytest.raises(ValueError, match="Key 'extra' without matching comment"):
        scen.add_set("i", ["i9", "extra"], ["i9 comment"])
    # Missing element in the index set
    with pytest.raises(
        ValueError, match="The index set 'i' does not have an " "element 'bar'!"
    ):
        scen.add_set("foo", "bar")

    # Retrieve set elements
    i = {"i1", "i2", "i3", "i4", "i5", "0", "1", "2", "i6", "i7", "i8"}
    assert i == set(scen.set("i"))

    # Remove elements from an 0D set: bare name
    scen.remove_set("i", "i2")
    i -= {"i2"}
    assert i == set(scen.set("i"))

    # Remove elements from an 0D set: Iterable of names, length >1
    scen.remove_set("i", ["i4", "i5"])
    i -= {"i4", "i5"}
    assert i == set(scen.set("i"))

    # Remove elements from a 1D set: Dict
    scen.remove_set("foo", dict(dim_i=["i7", "i8"]))
    # Added elements from above; minus directly removed elements; minus i2
    # because it was removed from the set i that indexes dim_i of foo
    foo = {"i1", "i2", "i3", "i6", "i7", "i8"} - {"i2"} - {"i7", "i8"}
    assert foo == set(scen.set("foo")["dim_i"])

    # Remove a set completely
    assert "h" not in scen.set_list()

    scen.init_set("h")
    assert "h" in scen.set_list()

    scen.remove_set("h")
    assert "h" not in scen.set_list()


def test_filter_str(scen_empty):
    scen = scen_empty

    elements = ["foo", 42, object()]
    expected = list(map(str, elements))

    scen.init_set("s")

    # Set elements can be added which are not str
    scen.add_set("s", elements)

    # Elements are stored and returned as str
    assert expected == scen.set("s").tolist()

    # Parameter defined over 's'
    p = pd.DataFrame.from_records(
        zip(elements, [1.0, 2.0, 3.0]), columns=["s", "value"]
    )

    # Expected return dtypes of index and value columns
    dtypes = {"s": str, "value": float}
    p_exp = p.astype(dtypes)

    scen.init_par("p", ["s"])
    scen.add_par("p", p)

    # Values can be retrieved using non-string filters
    exp = p_exp.loc[1:, :].reset_index(drop=True)
    obs = scen.par("p", filters={"s": elements[1:]})
    assert_frame_equal(exp[["s", "value"]], obs[["s", "value"]])


def test_solve_callback(test_mp):
    """Test the callback argument to Scenario.solve().

    In real usage, callback() would compute some kind of convergence criterion.
    This test uses a sequence of different values for d(seattle, new-york) in
    Dantzig's transport problem. Once the correct value is set on the
    ixmp.Scenario, the solution equals an expected value, and the model has
    'converged'.
    """
    # Set up the Dantzig problem
    scen = make_dantzig(test_mp)

    # Solve the scenario as configured
    solve_args = dict(model="dantzig", quiet=True)
    scen.solve(**solve_args)

    # Store the expected value of the decision variable, x
    expected = scen.var("x")

    # The reference distance between Seattle and New York is 2.5 [10^3 miles]
    d = [3.5, 2.0, 2.7, 2.5, 1.0]

    def set_d(scenario, value):
        """Set the distance between Seattle and New York to *value*."""
        scenario.remove_solution()
        scenario.check_out()
        data = {"i": "seattle", "j": "new-york", "value": value, "unit": "km"}
        # TODO should not be necessary here to call pd.DataFrame
        scenario.add_par("d", pd.DataFrame(data, index=[0]))
        scenario.commit("iterative solution")

    # Changing the entry in the array 'd' results in an optimal 'x' that is
    # different from the one stored as *expected*.
    set_d(scen, d[0])

    def change_distance(scenario):
        """Callback for model solution."""
        # Check if the model has 'converged' on the correct solution
        if (scenario.var("x") == expected).all(axis=None):
            return True

        # Convergence not reached

        # Change the distance between Seattle and New York, using the
        # 'iteration' variable stored on the Scenario object
        set_d(scenario, d[scenario.iteration])

        # commented: see below
        # # Trigger another solution of the model
        # return False

    # Warning is raised because 'return False' is commented above, meaning
    # user may have forgotten any return statement in the callback
    message = (
        r"solve\(callback=...\) argument returned None; will loop "
        "indefinitely unless True is returned."
    )
    with pytest.warns(UserWarning, match=message):
        # Model iterates automatically
        scen.solve(callback=change_distance, **solve_args)

    # Solution reached after 4 iterations, i.e. for d[4 - 1] == 2.5
    assert scen.iteration == 4



import logging
import os
import re
import shutil
import tempfile
from copy import copy
from pathlib import Path
from subprocess import CalledProcessError, check_call
from typing import Mapping

from ixmp.backend import ItemType
from ixmp.model.base import Model, ModelError
from ixmp.utils import as_str_list

log = logging.getLogger(__name__)


def gams_version():
    """Return the GAMS version as a string, e.g. '24.7.4'."""
    # NB check_output(['gams'], ...) does not work, because GAMS writes directly to the
    #    console instead of to stdout. check_output(['gams', '-LogOption=3'], ...) does
    #    not work, because GAMS does not accept options without an input file to
    #    execute.
    import os
    from subprocess import check_output
    from tempfile import mkdtemp

    # Create a temporary GAMS program that does nothing
    tmp_dir = Path(mkdtemp())
    gms = tmp_dir / "null.gms"
    gms.write_text("$exit;")

    # Execute, capturing stdout
    output = check_output(
        ["gams", "null", "-LogOption=3"],
        shell=os.name == "nt",
        cwd=tmp_dir,
        universal_newlines=True,
    )

    # Clean up
    gms.unlink()
    gms.with_suffix(".lst").unlink()
    tmp_dir.rmdir()

    # Find and return the version string
    pattern = r"^GAMS ([\d\.]+)\s*Copyright"
    return re.search(pattern, output, re.MULTILINE).groups()[0]


#: Return codes used by GAMS, from
#: https://www.gams.com/latest/docs/UG_GAMSReturnCodes.html . Values over 256 are only
#: valid on Windows, and are returned modulo 256 on other platforms.
RETURN_CODE = {
    0: "Normal return",
    1: "Solver is to be called, the system should never return this number",
    2: "There was a compilation error",
    3: "There was an execution error",
    4: "System limits were reached",
    5: "There was a file error",
    6: "There was a parameter error",
    7: "There was a licensing error",
    8: "There was a GAMS system error",
    9: "GAMS could not be started",
    10: "Out of memory",
    11: "Out of disk",
    109: "Could not create process/scratch directory",
    110: "Too many process/scratch directories",
    112: "Could not delete the process/scratch directory",
    113: "Could not write the script gamsnext",
    114: "Could not write the parameter file",
    115: "Could not read environment variable",
    400: "Could not spawn the GAMS language compiler (gamscmex)",
    401: "Current directory (curdir) does not exist",
    402: "Cannot set current directory (curdir)",
    404: "Blank in system directory",
    405: "Blank in current directory",
    406: "Blank in scratch extension (scrext)",
    407: "Unexpected cmexRC",
    408: "Could not find the process directory (procdir)",
    409: "CMEX library not be found (experimental)",
    410: "Entry point in CMEX library could not be found (experimental)",
    411: "Blank in process directory",
    412: "Blank in scratch directory",
    909: "Cannot add path / unknown UNIX environment / cannot set environment variable",
    1000: "Driver error: incorrect command line parameters for gams",
    2000: "Driver error: internal error: cannot install interrupt handler",
    3000: "Driver error: problems getting current directory",
    4000: "Driver error: internal error: GAMS compile and execute module not found",
    5000: "Driver error: internal error: cannot load option handling library",
}
RETURN_CODE = {key % 256: value for key, value in RETURN_CODE.items()}


class GAMSModel(Model):
    """General class for ixmp models using `GAMS <https://gams.com/>`_.

    GAMSModel solves a Scenario using the following steps:

    1. All Scenario data is written to a model input file in GDX format.
    2. A GAMS program is run to perform calculations, producing output in a
       GDX file.
    3. Output, or solution, data is read from the GDX file and stored in the
       Scenario.

    When created and :meth:`run`, GAMSModel constructs file paths and other
    necessary values using format strings. The :attr:`defaults` may be
    overridden by the keyword arguments to the constructor:

    Other parameters
    ----------------
    name : str, optional
        Override the :attr:`name` attribute to provide the `model_name` for
        format strings.
    model_file : str, optional
        Path to GAMS file, including '.gms' extension. Default: ``'{model_name}.gms'``
        in the current directory.
    case : str, optional
        Run or case identifier to use in GDX file names. Default:
        ``'{scenario.model}_{scenario.name}'``, where `scenario` is the
        :class:`.Scenario` object passed to :meth:`run`. Formatted using `model_name`
        and `scenario`.
    in_file : str, optional
        Path to write GDX input file. Default: ``'{model_name}_in.gdx'``. Formatted
        using `model_name`, `scenario`, and `case`.
    out_file : str, optional
        Path to read GDX output file. Default: ``'{model_name}_out.gdx'``. Formatted
        using `model_name`, `scenario`, and `case`.
    solve_args : list of str, optional
        Arguments to be passed to GAMS, e.g. to identify the model input and output
        files. Each formatted using `model_file`, `scenario`, `case`, `in_file`, and
        `out_file`. Default:

        - ``'--in="{in_file}"'``
        - ``'--out="{out_file}"'``
    gams_args : list of str, optional
        Additional arguments passed directly to GAMS without formatting, e.g. to
        control solver options or behaviour. See the `GAMS Documentation <https://www.gams.com/latest/docs/UG_GamsCall.html#UG_GamsCall_ListOfCommandLineParameters>`_.
        For example:

        - ``["iterLim=10"]`` limits the solver to 10 iterations.
    quiet: bool, optional
        If :obj:`True`, add "LogOption=2" to `gams_args` to redirect most console
        output during the model run to the log file. Default :obj:`False`, so
        "LogOption=4" is added. Any "LogOption" value provided explicitly via
        `gams_args` takes precedence.
    check_solution : bool, optional
        If :obj:`True`, raise an exception if the GAMS solver did not reach optimality.
        (Only for MESSAGE-scheme Scenarios.)
    comment : str, optional
        Comment added to Scenario when importing the solution. If omitted, no comment is
        added.
    equ_list : list of str, optional
        Equations to be imported from the `out_file`. Default: all.
    var_list : list of str, optional
        Variables to be imported from the `out_file`. Default: all.
    """  # noqa: E501

    #: Model name.
    name = "default"

    #: Default values and format strings for options.
    defaults: Mapping[str, object] = {
        "model_file": "{model_name}.gms",
        "case": "{scenario.model}_{scenario.scenario}",
        "in_file": str(Path("{cwd}", "{model_name}_in.gdx")),
        "out_file": str(Path("{cwd}", "{model_name}_out.gdx")),
        "solve_args": ['--in="{in_file}"', '--out="{out_file}"'],
        # Not formatted
        "gams_args": [],
        "check_solution": True,
        "comment": None,
        "equ_list": None,
        "var_list": None,
        "quiet": False,
        "use_temp_dir": True,
    }

    def __init__(self, name_=None, **model_options):
        self.model_name = self.clean_path(name_ or self.name)

        # Store options from `model_options`, otherwise from `defaults`
        for arg_name, default in self.defaults.items():
            setattr(self, arg_name, model_options.get(arg_name, copy(default)))

        # Check whether a subclass or user already set LogOption in `gams_args`
        if not any("LogOption" in arg for arg in self.gams_args):
            # Not set; use `quiet` to determine the value
            self.gams_args.append(f"LogOption={'2' if self.quiet else '4'}")

    def format_exception(self, exc, model_file):
        """Format a user-friendly exception when GAMS errors."""
        msg = [
            f"GAMS errored with return code {exc.returncode}:",
            # Convert a Windows return code >256 to its equivalent on *nix platforms
            f"    {RETURN_CODE[exc.returncode % 256]}",
            "",
            "For details, see the terminal output above, plus:",
            f"Input data: {self.in_file}",
        ]

        # Add a reference to the listing file, if it exists
        lst_file = Path(self.cwd).joinpath(model_file.name).with_suffix(".lst")
        if lst_file.exists():
            msg.insert(-1, f"Listing   : {lst_file}")

        return ModelError("\n".join(msg))

    def format_option(self, name):
        """Retrieve the option `name` and format it."""
        return self.format(getattr(self, name))

    def format(self, value):
        """Helper for recursive formatting of model options.

        `value` is formatted with replacements from the attributes of `self`.
        """
        try:
            return value.format(**self.__dict__)
        except AttributeError:
            # Something like a Path; don't format it
            return value

    def remove_temp_dir(self, msg="after run()"):
        """Remove the temporary directory, if any."""
        try:
            if self.use_temp_dir and self.cwd.exists():
                shutil.rmtree(self.cwd)
        except AttributeError:
            pass  # No .cwd, e.g. in a subclass
        except PermissionError as e:
            log.debug(f"Could not delete {repr(self.cwd)} {msg}")
            log.debug(str(e))

    def __del__(self):
        # Try once more to remove the temporary directory.
        # This appears to still fail on Windows.
        self.remove_temp_dir("at GAMSModel teardown")

    def run(self, scenario):
        """Execute the model."""
        # Store the scenario so its attributes can be referenced by format()
        self.scenario = scenario

        # Format or retrieve the model file option
        model_file = Path(self.format_option("model_file"))

        # Determine working directory for the GAMS call, possibly a temporary directory
        self.cwd = Path(tempfile.mkdtemp()) if self.use_temp_dir else model_file.parent
        # The "case" name
        self.case = self.clean_path(self.format_option("case").replace(" ", "_"))
        # Input and output file names
        self.in_file = Path(self.format_option("in_file"))
        self.out_file = Path(self.format_option("out_file"))

        # Assemble the full command: executable, model file, model-specific arguments,
        # and general GAMS arguments
        command = (
            ["gams", f'"{model_file}"']
            + [self.format(arg) for arg in self.solve_args]
            + self.gams_args
        )

        if os.name == "nt":
            # Windows: join the commands to a single string
            command = " ".join(command)

        # Remove stored reference to the Scenario to allow it to be GC'd later
        delattr(self, "scenario")

        # Common argument for write_file and read_file
        s_arg = dict(filters=dict(scenario=scenario))

        try:
            # Write model data to file
            scenario.platform._backend.write_file(
                self.in_file, ItemType.SET | ItemType.PAR, **s_arg
            )
        except NotImplementedError:  # pragma: no cover
            # No coverage because there currently is no such Backend that doesn't
            # support GDX

            # Remove the temporary directory, which should be empty
            self.remove_temp_dir()

            raise NotImplementedError(
                "GAMSModel requires a Backend that can write to GDX files, e.g. "
                "JDBCBackend"
            )

        try:
            # Invoke GAMS
            check_call(command, shell=os.name == "nt", cwd=self.cwd)
        except CalledProcessError as exc:
            # Do not remove self.temp_dir; the user may want to inspect the GDX file
            raise self.format_exception(exc, model_file) from None

        # Read model solution
        scenario.platform._backend.read_file(
            self.out_file,
            ItemType.MODEL,
            **s_arg,
            check_solution=self.check_solution,
            comment=self.comment or "",
            equ_list=as_str_list(self.equ_list) or [],
            var_list=as_str_list(self.var_list) or [],
        )

        # Finished: remove the temporary directory, if any
        self.remove_temp_dir()



"""Tests of :class:`ixmp.Platform`."""
import logging
import re
from sys import getrefcount
from weakref import getweakrefcount

import pandas as pd
import pytest
from pandas.testing import assert_frame_equal
from pytest import raises

import ixmp
from ixmp.backend import FIELDS
from ixmp.testing import DATA, assert_logs, models


def test_init():
    with pytest.raises(
        ValueError, match=re.escape("backend class 'foo' not among ['jdbc']")
    ):
        ixmp.Platform(backend="foo")


@pytest.fixture
def log_level_mp(test_mp):
    """A fixture that preserves the log level of *test_mp*."""
    tmp = test_mp.get_log_level()
    yield test_mp
    test_mp.set_log_level(tmp)


@pytest.mark.parametrize(
    "level, exc",
    [
        ("CRITICAL", None),
        ("ERROR", None),
        ("WARNING", None),
        ("INFO", None),
        ("DEBUG", None),
        ("NOTSET", None),
        # An unknown string fails
        ("FOO", ValueError),
    ],
)
def test_log_level(log_level_mp, level, exc):
    """Log level can be set and retrieved."""
    if exc is None:
        log_level_mp.set_log_level(level)
        assert log_level_mp.get_log_level() == level
    else:
        with pytest.raises(exc):
            log_level_mp.set_log_level(level)


def test_scenario_list(mp):
    scenario = mp.scenario_list(model="Douglas Adams")["scenario"]
    assert scenario[0] == "Hitchhiker"


def test_export_timeseries_data(test_mp, tmp_path):
    path = tmp_path / "export.csv"
    test_mp.export_timeseries_data(
        path, model="Douglas Adams", unit="???", region="World"
    )

    obs = pd.read_csv(path, index_col=False, header=0)

    exp = (
        DATA[0]
        .assign(**models["h2g2"], version=1, subannual="Year", meta=0)
        .rename(columns=lambda c: c.upper())
        .reindex(columns=FIELDS["write_file"])
    )

    assert_frame_equal(exp, obs)


def test_export_ts_wrong_params(test_mp, tmp_path):
    """Platform.export_timeseries_data to raise error with wrong parameters."""
    path = tmp_path / "export.csv"
    with raises(ValueError, match="Invalid arguments"):
        test_mp.export_timeseries_data(
            path,
            model="Douglas Adams",
            unit="???",
            region="World",
            export_all_runs=True,
        )


def test_export_ts_of_all_runs(mp, tmp_path):
    """Export timeseries of all runs."""
    path = tmp_path / "export.csv"

    # Add a new version of a run
    ts = ixmp.TimeSeries(mp, **models["h2g2"], version="new", annotation="fo")
    ts.add_timeseries(DATA[0])
    ts.commit("create a new version")
    ts.set_as_default()

    # Export all default model+scenario runs
    mp.export_timeseries_data(
        path, unit="???", region="World", default=True, export_all_runs=True
    )

    obs = pd.read_csv(path, index_col=False, header=0)
    exp = (
        DATA[0]
        .assign(**models["h2g2"], version=2, subannual="Year", meta=0)
        .rename(columns=lambda c: c.upper())
        .reindex(columns=FIELDS["write_file"])
    )

    assert_frame_equal(exp, obs)

    # Export all model+scenario run versions (including non-default)
    mp.export_timeseries_data(
        path, unit="???", region="World", default=False, export_all_runs=True
    )
    obs = pd.read_csv(path, index_col=False, header=0)
    assert 4 == len(obs)


def test_export_timeseries_data_empty(mp, tmp_path):
    """Dont export data if given models/scenarios do not have any runs."""
    path = tmp_path / "export.csv"
    model = "model-no-run"
    mp.add_model_name(model)
    mp.add_scenario_name("scenario-no-run")

    mp.export_timeseries_data(path, model=model, unit="???", region="World")

    assert 0 == len(pd.read_csv(path, index_col=False, header=0))


def test_unit_list(test_mp):
    units = test_mp.units()
    assert ("cases" in units) is True


def test_add_unit(test_mp):
    test_mp.add_unit("test", "just testing")


def test_regions(test_mp):
    regions = test_mp.regions()

    # Result has the expected columns
    columns = ["region", "mapped_to", "parent", "hierarchy"]
    assert all(regions.columns == columns)

    # One row is as expected
    obs = regions[regions.region == "World"]
    assert all([list(obs.loc[0]) == ["World", None, "World", "common"]])


def test_add_region(test_mp):
    # Region can be added
    test_mp.add_region("foo", "bar", "World")

    # Region can be retrieved
    regions = test_mp.regions()
    obs = regions[regions["region"] == "foo"].reset_index(drop=True)
    assert all([list(obs.loc[0]) == ["foo", None, "World", "bar"]])


def test_add_region_synonym(test_mp):
    test_mp.add_region("foo", "bar", "World")
    test_mp.add_region_synonym("foo2", "foo")
    regions = test_mp.regions()
    obs = regions[regions.region.isin(["foo", "foo2"])].reset_index(drop=True)

    exp = pd.DataFrame(
        [
            ["foo", None, "World", "bar"],
            ["foo2", "foo", "World", "bar"],
        ],
        columns=["region", "mapped_to", "parent", "hierarchy"],
    )
    assert_frame_equal(obs, exp)


def test_timeslices(test_mp):
    timeslices = test_mp.timeslices()
    obs = timeslices[timeslices.category == "Common"]
    # result has all attributes of time slice
    assert all(obs.columns == ["name", "category", "duration"])
    # result contains pre-defined YEAR time slice
    assert all([list(obs.iloc[0]) == ["Year", "Common", 1.0]])


def test_add_timeslice(test_mp):
    test_mp.add_timeslice("January, 1st", "Days", 1.0 / 366)
    timeslices = test_mp.timeslices()
    obs = timeslices[timeslices.category == "Days"]
    # return only added time slice
    assert len(obs) == 1
    # returned time slice attributes have expected values
    assert all([list(obs.iloc[0]) == ["January, 1st", "Days", 1.0 / 366]])


def test_add_timeslice_duplicate(caplog, test_mp):
    test_mp.add_timeslice("foo_slice", "foo_category", 0.2)

    # Adding same name with different duration raises an error
    msg = "timeslice `foo_slice` already defined with duration 0.2"
    with raises(ValueError, match=re.escape(msg)):
        test_mp.add_timeslice("foo_slice", "bar_category", 0.3)

    # Re-adding with the same duration only logs a message
    with assert_logs(caplog, msg, at_level=logging.INFO):
        test_mp.add_timeslice("foo_slice", "bar_category", 0.2)


def test_weakref():
    """Weak references allow Platforms to be del'd while Scenarios live."""
    mp = ixmp.Platform(
        backend="jdbc",
        driver="hsqldb",
        url="jdbc:hsqldb:mem:test_weakref",
    )

    # There is one reference to the Platform, and zero weak references
    assert getrefcount(mp) - 1 == 1
    assert getweakrefcount(mp) == 0

    # Create a single Scenario
    s = ixmp.Scenario(mp, "foo", "bar", version="new")

    # Still one reference to the Platform
    assert getrefcount(mp) - 1 == 1
    # …but additionally one weak reference
    assert getweakrefcount(mp) == 1

    # Make a local reference to the backend
    backend = mp._backend

    # Delete the Platform. Note that this only has an effect if there are no existing
    # references to it
    del mp

    # s.platform is a dead weak reference, so it can't be accessed
    with pytest.raises(ReferenceError):
        s.platform._backend

    # There is only one remaining reference to the backend: the *backend* name in the
    # local scope
    assert getrefcount(backend) - 1 == 1

    # The backend is garbage-collected at this point

    # The Scenario object still lives, but can't be used for anything
    assert s.model == "foo"

    # *s* is garbage-collected at this point


def test_add_model_name(test_mp):
    test_mp.add_model_name("new_model_name")
    assert "new_model_name" in test_mp.get_model_names()


def test_add_scenario_name(test_mp):
    test_mp.add_scenario_name("new_scenario_name")
    assert "new_scenario_name" in test_mp.get_scenario_names()



"""Tests for ixmp.utils."""
import logging

import numpy as np
import pandas as pd
import pandas.testing as pdt
import pytest
from pytest import mark, param

from ixmp import Scenario, utils
from ixmp.testing import make_dantzig, populate_test_platform


def test_check_year():

    # If y is a string value, raise a Value Error.

    y1 = "a"
    s1 = "a"
    with pytest.raises(ValueError):
        assert utils.check_year(y1, s1)

    # If y = None.

    y2 = None
    s2 = None

    assert utils.check_year(y2, s2) is None

    # If y is integer.

    y3 = 4
    s3 = 4

    assert utils.check_year(y3, s3) is True


def test_diff_identical(test_mp):
    """diff() of identical Scenarios."""
    scen_a = make_dantzig(test_mp)
    scen_b = make_dantzig(test_mp)

    # Compare identical scenarios: produces data of same length
    for name, df in utils.diff(scen_a, scen_b):
        data_a = utils.maybe_convert_scalar(scen_a.par(name))
        assert len(data_a) == len(df)

    # Compare identical scenarios, with filters
    iterator = utils.diff(scen_a, scen_b, filters=dict(i=["seattle"]))
    for (name, df), (exp_name, N) in zip(iterator, [("a", 1), ("d", 3)]):
        assert exp_name == name and len(df) == N


def test_diff_data(test_mp):
    """diff() when Scenarios contain the same items, but different data."""
    scen_a = make_dantzig(test_mp)
    scen_b = make_dantzig(test_mp)

    # Modify `scen_a` and `scen_b`
    scen_a.check_out()
    scen_b.check_out()

    # Remove elements from "b"
    drop_args = dict(labels=["value", "unit"], axis=1)
    scen_a.remove_par("b", scen_a.par("b").iloc[0:1, :].drop(**drop_args))
    scen_b.remove_par("b", scen_b.par("b").iloc[1:2, :].drop(**drop_args))
    # Remove elements from "d"
    scen_a.remove_par("d", scen_a.par("d").query("i == 'san-diego'").drop(**drop_args))
    # Modify values in "d"
    scen_b.add_par("d", scen_b.par("d").query("i == 'seattle'").assign(value=123.4))

    # Expected results
    exp_b = pd.DataFrame(
        [
            ["chicago", 300.0, "cases", np.NaN, None, "left_only"],
            ["new-york", np.NaN, None, 325.0, "cases", "right_only"],
            ["topeka", 275.0, "cases", 275.0, "cases", "both"],
        ],
        columns="j value_a unit_a value_b unit_b _merge".split(),
    )
    exp_d = pd.DataFrame(
        [
            ["san-diego", "chicago", np.NaN, None, 1.8, "km", "right_only"],
            ["san-diego", "new-york", np.NaN, None, 2.5, "km", "right_only"],
            ["san-diego", "topeka", np.NaN, None, 1.4, "km", "right_only"],
            ["seattle", "chicago", 1.7, "km", 123.4, "km", "both"],
            ["seattle", "new-york", 2.5, "km", 123.4, "km", "both"],
            ["seattle", "topeka", 1.8, "km", 123.4, "km", "both"],
        ],
        columns="i j value_a unit_a value_b unit_b _merge".split(),
    )

    # Use the specific categorical produced by pd.merge()
    merge_cat = pd.CategoricalDtype(["left_only", "right_only", "both"])
    exp_b = exp_b.astype(dict(_merge=merge_cat))
    exp_d = exp_d.astype(dict(_merge=merge_cat))

    # Compare different scenarios without filters
    for name, df in utils.diff(scen_a, scen_b):
        if name == "b":
            pdt.assert_frame_equal(exp_b, df)
        elif name == "d":
            pdt.assert_frame_equal(exp_d, df)

    # Compare different scenarios with filters
    iterator = utils.diff(scen_a, scen_b, filters=dict(j=["chicago"]))
    for name, df in iterator:
        # Same as above, except only the filtered rows should appear
        if name == "b":
            pdt.assert_frame_equal(exp_b.iloc[0:1, :], df)
        elif name == "d":
            pdt.assert_frame_equal(exp_d.iloc[[0, 3], :].reset_index(drop=True), df)


def test_diff_items(test_mp):
    """diff() when Scenarios contain the different items."""
    scen_a = make_dantzig(test_mp)
    scen_b = make_dantzig(test_mp)

    # Modify `scen_a` and `scen_b`
    scen_a.check_out()
    scen_b.check_out()

    # Remove items
    scen_a.remove_par("b")
    scen_b.remove_par("d")

    # Compare different scenarios without filters
    for name, df in utils.diff(scen_a, scen_b):
        pass  # No check on the contents

    # Compare different scenarios with filters
    iterator = utils.diff(scen_a, scen_b, filters=dict(j=["chicago"]))
    for name, df in iterator:
        pass  # No check of the contents


def test_logger_deprecated():
    with pytest.warns(DeprecationWarning):
        utils.logger()


m_s = dict(model="m", scenario="s")

URLS = [
    ("ixmp://example/m/s", dict(name="example"), m_s),
    (
        "ixmp://example/m/s#42",
        dict(name="example"),
        dict(model="m", scenario="s", version=42),
    ),
    ("ixmp://example/m/s", dict(name="example"), m_s),
    ("ixmp://local/m/s", dict(name="local"), m_s),
    (
        "ixmp://local/m/s/foo/bar",
        dict(name="local"),
        dict(model="m", scenario="s/foo/bar"),
    ),
    ("m/s#42", dict(), dict(model="m", scenario="s", version=42)),
    # Invalid values
    # Wrong scheme
    param("foo://example/m/s", None, None, marks=mark.xfail(raises=ValueError)),
    # No Scenario name
    param("ixmp://example/m", None, None, marks=mark.xfail(raises=ValueError)),
    # Version not an integer
    param(
        "ixmp://example/m#notaversion", None, None, marks=mark.xfail(raises=ValueError)
    ),
    # Query string not supported
    param(
        "ixmp://example/m/s?querystring",
        None,
        None,
        marks=mark.xfail(raises=ValueError),
    ),
]


@pytest.mark.parametrize("url, p, s", URLS)
def test_parse_url(url, p, s):
    platform_info, scenario_info = utils.parse_url(url)

    # Expected platform and scenario information is returned
    assert platform_info == p
    assert scenario_info == s


def test_format_scenario_list(test_mp_f):
    # Use the function-scoped fixture for precise version numbers
    mp = test_mp_f
    populate_test_platform(mp)

    # Expected results

    assert [
        "",
        "Douglas Adams/",
        "  Hitchhiker#1",
        "",
        "canning problem/",
        "  standard#2  1–3",
        "",
        "2 model name(s)",
        "2 scenario name(s)",
        "2 (model, scenario) combination(s)",
        "4 total scenarios",
    ] == utils.format_scenario_list(mp)

    # With as_url=True
    assert (
        list(
            map(
                lambda s: s.format(mp.name),
                [
                    "ixmp://{}/Douglas Adams/Hitchhiker#1",
                    "ixmp://{}/canning problem/standard#2",
                ],
            )
        )
        == utils.format_scenario_list(mp, as_url=True)
    )


def test_maybe_commit(caplog, test_mp):
    s = Scenario(test_mp, "maybe_commit", "maybe_commit", version="new")

    # A new Scenario is not committed, so this works
    assert utils.maybe_commit(s, True, message="foo") is True

    # *s* is already commited. No commit is performed, but the function call
    # succeeds and a message is logged
    caplog.set_level(logging.INFO, logger="ixmp")
    assert utils.maybe_commit(s, True, message="foo") is False
    assert caplog.messages[-1].startswith("maybe_commit() didn't commit: ")



import logging
from functools import partial

import pandas as pd
import pytest
from genno import Computer, Quantity
from genno.testing import assert_qty_equal
from pandas.testing import assert_frame_equal

from ixmp.model.dantzig import DATA
from ixmp.reporting.computations import map_as_qty, update_scenario
from ixmp.testing import assert_logs, make_dantzig

pytestmark = pytest.mark.usefixtures("parametrize_quantity_class")


def test_map_as_qty():
    b = ["b1", "b2", "b3", "b4"]
    input = pd.DataFrame(
        [["f1", "b1"], ["f1", "b2"], ["f2", "b3"]], columns=["foo", "bar"]
    )

    result = map_as_qty(input, b)

    exp = Quantity(
        pd.DataFrame(
            [
                ["f1", "b1", 1],
                ["f1", "b2", 1],
                ["f2", "b3", 1],
                ["all", "b1", 1],
                ["all", "b2", 1],
                ["all", "b3", 1],
                ["all", "b4", 1],
            ],
            columns=["foo", "bar", "value"],
        ).set_index(["foo", "bar"])
    )

    assert_qty_equal(exp, result)


def test_update_scenario(request, caplog, test_mp):
    scen = make_dantzig(test_mp)
    scen.check_out()
    scen.add_set("j", "toronto")
    scen.commit("Add j=toronto")

    # Number of rows in the 'd' parameter
    N_before = len(scen.par("d"))
    assert 6 == N_before

    # A Computer used as calculation engine
    c = Computer()

    # Target Scenario for updating data
    c.add("target", scen)

    # Create a pd.DataFrame suitable for Scenario.add_par()
    data = DATA["d"].query("j == 'chicago'").assign(j="toronto")
    data["value"] += 1.0

    # Add to the Reporter
    c.add("input", data)

    # Task to update the scenario with the data
    c.add("test 1", (partial(update_scenario, params=["d"]), "target", "input"))

    # Trigger the computation that results in data being added
    with assert_logs(caplog, f"'d' ← {len(data)} rows", at_level=logging.INFO):
        # Returns nothing
        assert c.get("test 1") is None

    # Rows were added to the parameter
    assert len(scen.par("d")) == N_before + len(data)

    # Modify the data
    data = pd.concat([DATA["d"], data]).reset_index(drop=True)
    data["value"] *= 2.0

    # Convert to a Quantity object and re-add
    q = Quantity(data.set_index(["i", "j"])["value"], name="d", units="km")
    c.add("input", q)

    # Revise the task; the parameter name ('demand') is read from the Quantity
    c.add("test 2", (update_scenario, "target", "input"))

    # Trigger the computation
    with assert_logs(caplog, f"'d' ← {len(data)} rows", at_level=logging.INFO):
        c.get("test 2")

    # All the rows have been updated
    assert_frame_equal(scen.par("d"), data)



import json
from abc import ABC, abstractmethod
from copy import copy
from typing import Any, Dict, Generator, MutableMapping, Sequence

from ixmp.core import Scenario, TimeSeries

from . import ItemType
from .io import s_read_excel, s_write_excel, ts_read_file


class Backend(ABC):
    """Abstract base class for backends."""

    def __init__(self):
        """OPTIONAL: Initialize the backend."""

    def __call__(self, obj, method, *args, **kwargs):
        """Call the backend method *method* for *obj*.

        The class attribute obj._backend_prefix is used to determine a prefix
        for the method name, e.g. 'ts_{method}'.
        """
        return getattr(self, method)(obj, *args, **kwargs)

    # Platform methods

    @classmethod
    def handle_config(cls, args: Sequence, kwargs: MutableMapping) -> Dict[str, Any]:
        """OPTIONAL: Handle platform/backend config arguments.

        Returns a :class:`dict` to be stored in the configuration file. This
        :class:`dict` **must** be valid as keyword arguments to the :meth:`__init__` of
        a backend subclass.

        The default implementation expects both `args` and `kwargs` to be empty.

        See also
        --------
        Config.add_platform
        """
        msg = "Unhandled {} args to Backend.handle_config(): {!r}"
        if len(args):
            raise ValueError(msg.format("positional", args))
        if len(kwargs):
            raise ValueError(msg.format("keyword", kwargs))
        return dict()

    def set_log_level(self, level):
        """OPTIONAL: Set logging level for the backend and other code.

        The default implementation has no effect.

        Parameters
        ----------
        level : int or Python logging level

        See also
        --------
        get_log_level
        """

    def get_log_level(self):
        """OPTIONAL: Get logging level for the backend and other code.

        The default implementation has no effect.

        Returns
        -------
        str
            Name of a :py:ref:`Python logging level <levels>`.

        See also
        --------
        set_log_level
        """

    @abstractmethod
    def set_doc(self, domain, docs):
        """Save documentation to database

        Parameters
        ----------
        domain : str
            Documentation domain, e.g. model, scenario etc
        docs : dict or array of tuples
            Dictionary or tuple array containing mapping between name of domain
            object (e.g. model name) and string representing fragment
            of documentation
        """

    @abstractmethod
    def get_doc(self, domain, name=None):
        """Read documentation from database

        Parameters
        ----------
        domain : str
            Documentation domain, e.g. model, scenario etc
        name : str, optional
            Name of domain entity (e.g. model name).

        Returns
        -------
        str or dict
            String representing fragment of documentation if name is passed as
            parameter or dictionary containing mapping between name of domain
            object (e.g. model name) and string representing fragment when
            name parameter is omitted.
        """

    def open_db(self):
        """OPTIONAL: (Re-)open database connection(s).

        A backend **may** connect to a database server. This method opens the
        database connection if it is closed.
        """

    def close_db(self):
        """OPTIONAL: Close database connection(s).

        Close any database connection(s), if open.
        """

    def get_auth(self, user, models, kind):
        """OPTIONAL: Return user authorization for *models*.

        If the Backend implements access control, this method **must** indicate
        whether *user* has permission *kind* for each of *models*.

        *kind* **may** be 'read'/'view', 'write'/'modify', or other values;
        :meth:`get_auth` **should** raise exceptions on invalid values.

        Parameters
        ----------
        user : str
            User name or identifier.
        models : list of str
            Model names.
        kind : str
            Type of permission being requested

        Returns
        -------
        dict
            Mapping of *model name (str)* → :class:`bool`; :obj:`True` if the
            user is authorized for the model.
        """
        return {model: True for model in models}

    @abstractmethod
    def set_node(self, name, parent=None, hierarchy=None, synonym=None):
        """Add a node name to the Platform.

        This method **must** have one of two effects, depending on the
        arguments:

        - With `parent` and `hierarchy`: `name` is added as a child of `parent`
          in the named `hierarchy`.
        - With `synonym`: `synonym` is added as an alias for `name`.

        Parameters
        ----------
        name : str
           Node name.
        parent : str, optional
           Parent node name.
        hierarchy : str, optional
           Node hierarchy ID.
        synonym : str, optional
           Synonym for node.

        See also
        --------
        get_nodes
        """

    @abstractmethod
    def get_nodes(self):
        """Iterate over all nodes stored on the Platform.

        Yields
        -------
        tuple
            The members of each tuple are:

            ========= =========== ===
            ID        Type        Description
            ========= =========== ===
            region    str         Node name or synonym for node
            mapped_to str or None Node name
            parent    str         Parent node name
            hierarchy str         Node hierarchy ID
            ========= =========== ===

        See also
        --------
        set_node
        """

    @abstractmethod
    def get_timeslices(self):
        """Iterate over subannual timeslices defined on the Platform instance.

        Yields
        -------
        tuple
            The members of each tuple are:

            ========= =========== ===
            ID        Type        Description
            ========= =========== ===
            name      str         Time slice name
            category  str         Time slice category
            duration  float       Time slice duration (fraction of year)
            ========= =========== ===

        See also
        --------
        set_timeslice
        """

    @abstractmethod
    def set_timeslice(self, name, category, duration):
        """Add a subannual time slice to the Platform.

        Parameters
        ----------
        name : str
           Node name.
        category : str
           Time slice category.
        duration : float
           Time slice duration (a fraction of a year).

        See also
        --------
        get_timeslices
        """

    @abstractmethod
    def add_model_name(self, name: str):
        """Add (register) new model name.

        Parameters
        ----------
        name : str
            New model name
        """

    @abstractmethod
    def add_scenario_name(self, name: str):
        """Add (register) new scenario name.

        Parameters
        ----------
        name : str
            New scenario name
        """

    @abstractmethod
    def get_model_names(self) -> Generator[str, None, None]:
        """List existing model names.

        Returns
        -------
        list of str
            List of the retrieved model names.
        """

    @abstractmethod
    def get_scenario_names(self) -> Generator[str, None, None]:
        """List existing scenario names.

        Returns
        -------
        list of str
            List of the retrieved scenario names.
        """

    @abstractmethod
    def get_scenarios(self, default, model, scenario):
        """Iterate over TimeSeries stored on the Platform.

        Scenarios, as subclasses of TimeSeries, are also included.

        Parameters
        ----------
        default : bool
           :obj:`True` to include only TimeSeries versions marked as default.
        model : str or None
           Model name to filter results.
        scenario : str or None
           Scenario name to filter results.

        Yields
        ------
        tuple
            The members of each tuple are:

            ========== ==== ===
            ID         Type Description
            ========== ==== ===
            model      str  Model name
            scenario   str  Scenario name
            scheme     str  Scheme name
            is_default bool :obj:`True` if `version` is the default
            is_locked  bool :obj:`True` if read-only
            cre_user   str  Name of user who created the TimeSeries
            cre_date   str  Creation datetime
            upd_user   str  Name of user who last modified the TimeSeries
            upd_date   str  Modification datetime
            lock_user  str  Name of user who locked the TimeSeries
            lock_date  str  Lock datetime
            annotation str  Description of the TimeSeries
            version    int  Version
            ========== ==== ===
        """

    @abstractmethod
    def set_unit(self, name, comment):
        """Add a unit of measurement to the Platform.

        Parameters
        ----------
        name : str
            Symbol of the unit.
        comment : str
            Description of the change or of the unit.

        See also
        --------
        get_units
        """

    @abstractmethod
    def get_units(self):
        """Return all registered symbols for units of measurement.

        Returns
        -------
        list of str

        See also
        --------
        set_unit
        """

    def read_file(self, path, item_type: ItemType, **kwargs):
        """OPTIONAL: Read Platform, TimeSeries, or Scenario data from file.

        A backend **may** implement read_file for one or more combinations of
        the `path` and `item_type` methods. For all other combinations, it
        **must** raise :class:`NotImplementedError`.

        The default implementation supports:

        - `path` ending in '.xlsx', `item_type` is ItemType.MODEL: read a
          single Scenario given by kwargs['filters']['scenario'] from file
          using :meth:`pandas.DataFrame.read_excel`.

        Parameters
        ----------
        path : os.PathLike
            File for input. The filename suffix determines the input format:

            ====== ===
            Suffix Format
            ====== ===
            .csv   Comma-separated values
            .gdx   GAMS data exchange
            .xlsx  Microsoft Office Open XML spreadsheet
            ====== ===

        item_type : ItemType
            Type(s) of items to read.

        Raises
        ------
        ValueError
            If `ts` is not None and 'scenario' is a key in `filters`.
        NotImplementedError
            If input of the specified items from the file format is not
            supported.

        See also
        --------
        write_file
        """
        s, filters = self._handle_rw_filters(kwargs.pop("filters", {}))
        if path.suffix in (".csv", ".xlsx") and item_type is ItemType.TS and s:
            ts_read_file(s, path, **kwargs)
        elif path.suffix == ".xlsx" and item_type is ItemType.MODEL and s:
            s_read_excel(self, s, path, **kwargs)
        else:
            raise NotImplementedError

    def write_file(self, path, item_type: ItemType, **kwargs):
        """OPTIONAL: Write Platform, TimeSeries, or Scenario data to file.

        A backend **may** implement write_file for one or more combinations of
        the `path` and `item_type` methods. For all other combinations, it
        **must** raise :class:`NotImplementedError`.

        The default implementation supports:

        - `path` ending in '.xlsx', `item_type` is either :attr:`.MODEL` or
          :attr:`.SET` | :attr:`.PAR`: write a single Scenario given by
          kwargs['filters']['scenario'] to file using
          :meth:`pandas.DataFrame.to_excel`.

        Parameters
        ----------
        path : os.PathLike
            File for output. The filename suffix determines the output format.
        item_type : ItemType
            Type(s) of items to write.

        Raises
        ------
        ValueError
            If `ts` is not None and 'scenario' is a key in `filters`.
        NotImplementedError
            If output of the specified items to the file format is not
            supported.

        See also
        --------
        read_file
        """
        # Use the "scenario" filter to retrieve the Scenario `s` to be written;
        # reappend any other filters
        s, kwargs["filters"] = self._handle_rw_filters(kwargs.pop("filters", {}))

        xlsx_types = (ItemType.SET | ItemType.PAR, ItemType.MODEL)
        if path.suffix == ".xlsx" and item_type in xlsx_types and s:
            s_write_excel(self, s, path, item_type, **kwargs)
        else:
            # All other combinations of arguments
            raise NotImplementedError

    # Methods for ixmp.TimeSeries

    @abstractmethod
    def init(self, ts: TimeSeries, annotation):
        """Create a new TimeSeries (or Scenario) *ts*.

        init **may** modify the :attr:`~TimeSeries.version` attribute of
        *ts*.

        If *ts* is a :class:`Scenario`; the Backend **must** store the
        :attr:`.Scenario.scheme` attribute.

        Parameters
        ----------
        annotation : str
            If *ts* is newly-created, the Backend **must** store this
            annotation with the TimeSeries.

        Returns
        -------
        None
        """

    @abstractmethod
    def get(self, ts: TimeSeries):
        """Retrieve the existing TimeSeries (or Scenario) *ts*.

        The TimeSeries is identified based on the unique combination of the
        attributes of *ts*:

        - :attr:`~.TimeSeries.model`,
        - :attr:`~.TimeSeries.scenario`, and
        - :attr:`~.TimeSeries.version`.

        If :attr:`.version` is :obj:`None`, the Backend **must** return the
        version marked as default, and **must** set the attribute value.

        If *ts* is a Scenario, :meth:`get` **must** set the
        :attr:`~.Scenario.scheme` attribute with the value previously passed to
        :meth:`init`.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If :attr:`~.TimeSeries.model` or :attr:`~.TimeSeries.scenario` does
            not exist on the Platform.

        See also
        --------
        is_default
        set_as_default
        """

    def del_ts(self, ts: TimeSeries):
        """OPTIONAL: Free memory associated with the TimeSeries *ts*.

        The default implementation has no effect.
        """

    @abstractmethod
    def check_out(self, ts: TimeSeries, timeseries_only):
        """Check out *ts* for modification.

        Parameters
        ----------
        timeseries_only : bool
            ???

        Returns
        -------
        None
        """

    @abstractmethod
    def commit(self, ts: TimeSeries, comment):
        """Commit changes to *ts*.

        ts_init **may** modify the :attr:`~.TimeSeries.version` attribute of
        *ts*.

        Parameters
        ----------
        comment : str
            Description of the changes being committed.

        Returns
        -------
        None
        """

    @abstractmethod
    def discard_changes(self, ts: TimeSeries):
        """Discard changes to *ts* since the last :meth:`ts_check_out`.

        Returns
        -------
        None
        """

    @abstractmethod
    def set_as_default(self, ts: TimeSeries):
        """Set the current :attr:`.TimeSeries.version` as the default.

        Returns
        -------
        None

        See also
        --------
        get
        is_default
        """

    @abstractmethod
    def is_default(self, ts: TimeSeries):
        """Return :obj:`True` if *ts* is the default version.

        Returns
        -------
        bool

        See also
        --------
        get
        set_as_default
        """

    @abstractmethod
    def last_update(self, ts: TimeSeries):
        """Return the date of the last modification of the *ts*.

        Returns
        -------
        str or None
        """

    @abstractmethod
    def run_id(self, ts: TimeSeries):
        """Return the run ID for the *ts*.

        Returns
        -------
        int
        """

    def preload(self, ts: TimeSeries):
        """OPTIONAL: Load *ts* data into memory."""

    @staticmethod
    def _handle_rw_filters(filters: dict):
        """Helper for :meth:`read_file` and :meth:`write_file`.

        The `filters` argument is unpacked if the 'scenarios' key is a single
        :class:`TimeSeries` object. A 2-tuple is returned of the object (or
        :obj:`None`) and the remaining filters.
        """
        ts = None
        filters = copy(filters)
        try:
            if isinstance(filters["scenario"], TimeSeries):
                ts = filters.pop("scenario")
        except KeyError:
            pass  # Don't modify filters at all

        return ts, filters

    @abstractmethod
    def get_data(self, ts: TimeSeries, region, variable, unit, year):
        """Retrieve time-series data.

        Parameters
        ----------
        region : list of str
            Region names to filter results.
        variable : list of str
            Variable names to filter results.
        unit : list of str
            Unit symbols to filter results.
        year : list of str
            Years to filter results.

        Yields
        ------
        tuple
            The members of each tuple are:

            ======== ===== ===
            ID       Type  Description
            ======== ===== ===
            region   str   Region name
            variable str   Variable name
            unit     str   Unit symbol
            year     int   Year
            value    float Data value
            ======== ===== ===
        """

    @abstractmethod
    def get_geo(self, ts: TimeSeries):
        """Retrieve time-series 'geodata'.

        Yields
        ------
        tuple
            The members of each tuple are:

            ========= ==== ===
            ID        Type Description
            ========= ==== ===
            region    str  Region name
            variable  str  Variable name
            year      int  Year
            value     str  Value
            unit      str  Unit symbol
            subannual str  Name of time slice
            meta      bool :obj:`True` if the data is marked as metadata
            ========= ==== ===
        """

    @abstractmethod
    def set_data(self, ts: TimeSeries, region, variable, data, unit, subannual, meta):
        """Store *data*.

        Parameters
        ----------
        region : str
            Region name.
        variable : str
            Variable name.
        subannual : str
            Name of time slice.
        unit : str
            Unit symbol.
        data : dict (int -> float)
            Mapping from year to value.
        meta : bool
            :obj:`True` to mark *data* as metadata.
        """

    @abstractmethod
    def set_geo(
        self, ts: TimeSeries, region, variable, subannual, year, value, unit, meta
    ):
        """Store time-series 'geodata'.

        Parameters
        ----------
        region : str
            Region name.
        variable : str
            Variable name.
        subannual : str
            Name of time slice.
        year : int
            Year.
        value : str
            Data value.
        unit : str
            Unit symbol.
        meta : bool
            :obj:`True` to mark *data* as metadata.
        """

    @abstractmethod
    def delete(self, ts: TimeSeries, region, variable, subannual, years, unit):
        """Remove data values.

        Parameters
        ----------
        region : str
            Region name.
        variable : str
            Variable name.
        years : Iterable of int
            Years.
        unit : str
            Unit symbol.
        subannual : str
            Name of time slice.

        Returns
        -------
        None
        """

    @abstractmethod
    def delete_geo(self, ts: TimeSeries, region, variable, subannual, years, unit):
        """Remove 'geodata' values.

        Parameters
        ----------
        region : str
            Region name.
        variable : str
            Variable name.
        subannual : str
            Name of time slice.
        years : Iterable of int
            Years.
        unit : str
            Unit symbol.

        Returns
        -------
        None
        """

    # Methods for ixmp.Scenario

    @abstractmethod
    def clone(
        self,
        s: Scenario,
        platform_dest,
        model,
        scenario,
        annotation,
        keep_solution,
        first_model_year=None,
    ):
        """Clone *s*.

        Parameters
        ----------
        platform_dest : :class:`.Platform`
            Target backend. May be the same as *s.platform*.
        model : str
            New model name.
        scenario : str
            New scenario name.
        annotation : str
            Description for the creation of the new scenario.
        keep_solution : bool
            If :obj:`True`, model solution data is also cloned. If
            :obj:`False`, it is discarded.
        first_model_year : int or None
            If :class:`int`, must be greater than the first model year of *s*.

        Returns
        -------
        Same class as *s*
            The cloned Scenario.
        """

    @abstractmethod
    def has_solution(self, s: Scenario):
        """Return `True` if Scenario *s* has been solved.

        If :obj:`True`, model solution data is available from the Backend.
        """

    @abstractmethod
    def list_items(self, s: Scenario, type):
        """Return a list of items of *type*.

        Parameters
        ----------
        type : 'set' or 'par' or 'equ'

        Return
        ------
        list of str
        """

    @abstractmethod
    def init_item(self, s: Scenario, type, name, idx_sets, idx_names):
        """Initialize an item *name* of *type*.

        Parameters
        ----------
        type : 'set' or 'par' or 'equ' or 'var'
        name : str
            Name for the new item.
        idx_sets : sequence of str
            If empty, a 0-dimensional/scalar item is initialized. Otherwise, a
            1+-dimensional item is initialized.
        idx_names : sequence of str or None
            Optional names for the dimensions. If not supplied, the names of
            the *idx_sets* (if any) are used. If supplied, *idx_names* and
            *idx_sets* must be the same length.

        Return
        ------
        None

        Raises
        ------
        ValueError
            if any of the *idx_sets* is not an existing set in the Scenario;
            if *idx_names* and *idx_sets* are not the same length.
        """

    @abstractmethod
    def delete_item(self, s: Scenario, type, name):
        """Remove an item *name* of *type*.

        Parameters
        ----------
        type : 'set' or 'par' or 'equ'
        name : str
            Name of the item to delete.

        Return
        ------
        None
        """

    @abstractmethod
    def item_index(self, s: Scenario, name, sets_or_names):
        """Return the index sets or names of item *name*.

        Parameters
        ----------
        sets_or_names : 'sets' or 'names'

        Returns
        -------
        list of str
        """

    @abstractmethod
    def item_get_elements(self, s: Scenario, type, name, filters=None):
        """Return elements of item *name*.

        Parameters
        ----------
        type : 'equ' or 'par' or 'set' or 'var'
        name : str
            Name of the item.
        filters : dict (str -> list), optional
            If provided, a mapping from dimension names to allowed values
            along that dimension.

            item_get_elements **must** silently accept values that are *not*
            members of the set indexing a dimension. Elements which are not
            :class:`str` **must** be handled as equivalent to their string
            representation; i.e. item_get_elements must return the same data
            for `filters={'foo': [42]}` and `filters={'foo': ['42']}`.

        Returns
        -------
        pandas.Series
            When *type* is 'set' and *name* an index set (not indexed by other
            sets).
        dict
            When *type* is 'equ', 'par', or 'var' and *name* is scalar (zero-
            dimensional). The value has the keys 'value' and 'unit' (for 'par')
            or 'lvl' and 'mrg' (for 'equ' or 'var').
        pandas.DataFrame
            For mapping sets, or all 1+-dimensional values. The dataframe has
            one column per index name with dimension values; plus the columns
            'value' and 'unit' (for 'par') or 'lvl' and 'mrg' (for 'equ' or
            'var').

        Raises
        ------
        KeyError
            If *name* does not exist in *s*.
        """

    @abstractmethod
    def item_set_elements(self, s: Scenario, type, name, elements):
        """Add keys or values to item *name*.

        Parameters
        ----------
        type : 'par' or 'set'
        name : str
            Name of the items.
        elements : iterable of 4-tuple
            The members of each tuple are:

            ======= ========================== ===
            ID      Type                       Description
            ======= ========================== ===
            key     str or list of str or None Set elements or value indices
            value   float or None              Parameter value
            unit    str or None                Unit symbol
            comment str or None                Description of the change
            ======= ========================== ===

            If *name* is indexed by other set(s), then the number of elements
            of each *key*, and their contents, must match the index set(s).
            When *type* is 'set', *value* and *unit* **must** be :obj:`None`.

        Raises
        ------
        ValueError
            If *elements* contain invalid values, e.g. key values not in the
            index set(s).
        Exception
            If the Backend encounters any error adding the elements.

        See also
        --------
        s_init_item
        s_item_delete_elements
        """

    @abstractmethod
    def item_delete_elements(self, s: Scenario, type, name, keys):
        """Remove elements of item *name*.

        Parameters
        ----------
        type : 'par' or 'set'
        name : str
        keys : iterable of iterable of str
            If *name* is indexed by other set(s), then the number of elements
            of each key in *keys*, and their contents, must match the index
            set(s).
            If *name* is a basic set, then each key must be a list containing a
            single str, which must exist in the set.

        See also
        --------
        s_init_item
        s_item_set_elements
        """

    @abstractmethod
    def get_meta(self, model: str, scenario: str, version: int, strict: bool) -> dict:
        """Retrieve meta indicators.

        Parameters
        ----------
        model : str, optional
            filter meta by a model
        scenario : str, optional
            filter meta by a scenario
        version : int or str, optional
            retrieve meta of a specific model/scenario run version
        strict : bool, optional
            only retrieve indicators from the requested model-scenario-version
            level

        Returns
        -------
        dict (str -> any)
            Mapping from meta category keys to values.

        Raises
        ------
        ValueError
            On unsupported model-scenario-version combinations.
            Supported combinations are: (model), (scenario), (model, scenario),
            (model, scenario, version)
        """

    @abstractmethod
    def set_meta(self, meta: dict, model: str, scenario: str, version: int):
        """Set meta categories.

        Parameters
        ----------
        meta : dict
            containing meta key/value category pairs
        model : str, optional
            model name that meta should be attached to
        scenario : str, optional
            scenario name that meta should be attached to
        version : int, optional
            run version that meta should be attached to

        Returns
        -------
        None

        Raises
        ------
        ValueError
            On unsupported model-scenario-version combinations.
            Supported combinations are: (model), (scenario), (model, scenario),
            (model, scenario, version)
        """

    @abstractmethod
    def remove_meta(self, categories: list, model: str, scenario: str, version: int):
        """Remove meta categories.

        Parameters
        ----------
        categories : list of str
            meta-category keys to remove
        model : str, optional
            only remove meta of a specific model
        scenario : str, optional
            only remove meta of a specific scenario
        version : int, optional
            only remove meta of a specific model/scenario run version

        Returns
        -------
        None

        Raises
        ------
        ValueError
            On unsupported model-scenario-version combinations.
            Supported combinations are: (model), (scenario), (model, scenario),
            (model, scenario, version)
        """

    @abstractmethod
    def clear_solution(self, s: Scenario, from_year=None):
        """Remove data associated with a model solution.

        .. todo:: Document.
        """

    # Methods for message_ix.Scenario

    @abstractmethod
    def cat_list(self, ms: Scenario, name):
        """Return list of categories in mapping *name*.

        Parameters
        ----------
        name : str
            Name of the category mapping set.

        Returns
        -------
        list of str
            All categories in *name*.
        """

    @abstractmethod
    def cat_get_elements(self, ms: Scenario, name, cat):
        """Get elements of a category mapping.

        Parameters
        ----------
        name : str
            Name of the category mapping set.
        cat : str
            Name of the category within *name*.

        Returns
        -------
        list of str
            All set elements mapped to *cat* in *name*.
        """

    @abstractmethod
    def cat_set_elements(self, ms: Scenario, name, cat, keys, is_unique):
        """Add elements to category mapping.

        Parameters
        ----------
        name : str
            Name of the category mapping set.
        cat : str
            Name of the category within *name*.
        keys : iterable of str or list of str
            Keys to add to *cat*.
        is_unique : bool
            If :obj:`True`:

            - *keys* **must** contain only one key.
            - The Backend **must** remove any existing member of *cat*, so that
              it has only one element.

        Returns
        -------
        None
        """


class CachingBackend(Backend):
    """Backend with additional features for caching data."""

    #: :obj:`True` if caching is enabled.
    cache_enabled = True

    #: Cache of values. Keys are given by :meth:`_cache_key`; values depend on
    #: the subclass' usage of the cache.
    _cache: Dict[str, object] = {}

    #: Count of number of times a value was retrieved from cache successfully
    #: using :meth:`cache_get`.
    _cache_hit: Dict[str, int] = {}

    # Backend API methods

    def __init__(self, cache_enabled=True):
        super().__init__()

        self.cache_enabled = cache_enabled

        # Empty the cache
        self._cache = {}
        self._cache_hit = {}

    def del_ts(self, ts: TimeSeries):
        """Invalidate cache entries associated with *ts*."""
        self.cache_invalidate(ts)

    # New methods for CachingBackend

    @classmethod
    def _cache_key(self, ts, ix_type, name, filters=None):
        """Return a hashable cache key.

        ixmp *filters* (a :class:`dict` of :class:`list`) are converted to a
        unique id that is hashable.

        Parameters
        ----------
        ts : .TimeSeries
        ix_type : str
        name : str
        filters : dict (str -> list of hashable)

        Returns
        -------
        tuple
            A hashable key with 4 elements for *ts*, *ix_type*, *name*, and
            *filters*.
        """
        ts = id(ts)
        if filters is None or len(filters) == 0:
            return (ts, ix_type, name)
        else:
            # Convert filters into a hashable object
            filters = hash(json.dumps(sorted(filters.items())))
            return (ts, ix_type, name, filters)

    def cache_get(self, ts, ix_type, name, filters):
        """Retrieve value from cache.

        The value in :attr:`_cache` is copied to avoid cached values being
        modified by user code. :attr:`_cache_hit` is incremented.

        Raises
        ------
        KeyError
            If the key for *ts*, *ix_type*, *name* and *filters* is not in the
            cache.
        """
        key = self._cache_key(ts, ix_type, name, filters)

        if self.cache_enabled and key in self._cache:
            self._cache_hit[key] = self._cache_hit.setdefault(key, 0) + 1
            return copy(self._cache[key])
        else:
            raise KeyError(ts, ix_type, name, filters)

    def cache(self, ts, ix_type, name, filters, value):
        """Store *value* in cache.

        Returns
        -------
        bool
            :obj:`True` if the key was already in the cache and its value was
            overwritten.
        """
        if not self.cache_enabled:
            # Don't store anything if cache is disabled
            return False

        key = self._cache_key(ts, ix_type, name, filters)

        refreshed = key in self._cache
        self._cache[key] = value

        return refreshed

    def cache_invalidate(self, ts, ix_type=None, name=None, filters=None):
        """Invalidate cached values.

        With all arguments given, single key/value is removed from the cache.
        Otherwise, multiple keys/values are removed:

        - *ts* only: all cached values associated with the :class:`.TimeSeries`
          or :class:`.Scenario` object.
        - *ts*, *ix_type*, and *name*: all cached values associated with the
          ixmp item, whether filtered or unfiltered.
        """
        key = self._cache_key(ts, ix_type, name, filters)

        if filters is None:
            i = slice(1) if (ix_type is name is None) else slice(3)
            to_remove = filter(lambda k: k[i] == key[i], self._cache.keys())
        else:
            to_remove = [key]

        for key in list(to_remove):
            self._cache.pop(key)



from enum import IntFlag
from typing import Dict, Type

#: Lists of field names for tuples returned by Backend API methods.
#:
#: The key "write_file" refers to the columns appearing in the CSV output from
#: :meth:`export_timeseries_data` when using :class:`.JDBCBackend`.
#:
#: .. todo:: Make this consistent with other dimension orders and with :data:`IAMC_IDX`.
FIELDS = {
    "get_nodes": ("region", "mapped_to", "parent", "hierarchy"),
    "get_timeslices": ("name", "category", "duration"),
    "get_scenarios": (
        "model",
        "scenario",
        "scheme",
        "is_default",
        "is_locked",
        "cre_user",
        "cre_date",
        "upd_user",
        "upd_date",
        "lock_user",
        "lock_date",
        "annotation",
        "version",
    ),
    "ts_get": ("region", "variable", "unit", "subannual", "year", "value"),
    "ts_get_geo": ("region", "variable", "subannual", "year", "value", "unit", "meta"),
    "write_file": (
        "MODEL",
        "SCENARIO",
        "VERSION",
        "VARIABLE",
        "UNIT",
        "REGION",
        "META",
        "SUBANNUAL",
        "YEAR",
        "VALUE",
    ),
}


#: Mapping from names to available backends. To register additional backends,
#: add elements to this variable.
BACKENDS: Dict[str, Type] = {}


class ItemType(IntFlag):
    """Type of data items in :class:`.TimeSeries` and :class:`.Scenario`."""

    # NB the docstring comments ('#:') are placed as they are to ensure the
    #    output is readable.

    TS = 1
    #: Time series data variable.
    T = TS

    SET = 2
    #: Set.
    S = SET

    PAR = 4
    #: Parameter.
    P = PAR

    VAR = 8
    #: Model variable.
    V = VAR

    EQU = 16
    #: Equation.
    E = EQU

    MODEL = SET + PAR + VAR + EQU
    #: All kinds of model-related data, i.e. :attr:`SET`, :attr:`PAR`,
    #: :attr:`VAR` and :attr:`EQU`.
    M = MODEL

    #: Model solution data, i.e. :attr:`VAR` and :attr:`EQU`.
    SOLUTION = VAR + EQU

    ALL = TS + MODEL
    #: All data, i.e. :attr:`MODEL` and :attr:`TS`.
    A = ALL



import logging
import re

import pint
import pytest
from genno import ComputationError, configure

import ixmp
from ixmp.reporting.reporter import Reporter
from ixmp.reporting.util import RENAME_DIMS
from ixmp.testing import add_test_data, assert_logs, make_dantzig

pytestmark = pytest.mark.usefixtures("parametrize_quantity_class")

test_args = ("Douglas Adams", "Hitchhiker")


@pytest.fixture
def scenario(test_mp):
    # from test_feature_timeseries.test_new_timeseries_as_year_value
    scen = ixmp.Scenario(test_mp, *test_args, version="new", annotation="foo")
    scen.commit("importing a testing timeseries")
    yield scen


def test_configure(test_mp, test_data_path):
    # Configure globally; reads 'rename_dims' section
    configure(rename_dims={"i": "i_renamed"})

    # Reporting uses the RENAME_DIMS mapping of 'i' to 'i_renamed'
    scen = make_dantzig(test_mp)
    rep = Reporter.from_scenario(scen)
    assert "d:i_renamed-j" in rep, rep.graph.keys()
    assert ["seattle", "san-diego"] == rep.get("i_renamed")

    # Original name 'i' are not found in the reporter
    assert "d:i-j" not in rep, rep.graph.keys()
    pytest.raises(KeyError, rep.get, "i")

    # Remove the configuration for renaming 'i', so that other tests work
    RENAME_DIMS.pop("i")


def test_reporter_from_scenario(scenario):
    r = Reporter.from_scenario(scenario)

    r.finalize(scenario)

    assert "scenario" in r.graph


def test_platform_units(test_mp, caplog, ureg):
    """Test handling of units from ixmp.Platform.

    test_mp is loaded with some units including '-', '???', 'G$', etc. which
    are not parseable with pint; and others which are not defined in a default
    pint.UnitRegistry. These tests check the handling of those units.
    """

    # Prepare a Scenario with test data
    scen = ixmp.Scenario(
        test_mp, "reporting_platform_units", "reporting_platform_units", "new"
    )
    t, t_foo, t_bar, x = add_test_data(scen)
    rep = Reporter.from_scenario(scen)
    x_key = rep.full_key("x")

    # Convert 'x' to dataframe
    x = x.to_series().rename("value").reset_index()

    # Exception message, formatted as a regular expression
    msg = r"unit '{}' cannot be parsed; contains invalid character\(s\) '{}'"

    # Unit and components for the regex
    bad_units = [("-", "-", "-"), ("???", r"\?\?\?", r"\?"), ("E$", r"E\$", r"\$")]
    for unit, expr, chars in bad_units:
        # Add the unit
        test_mp.add_unit(unit)

        # Overwrite the parameter
        x["unit"] = unit
        scen.add_par("x", x)

        # Parsing units with invalid chars raises an intelligible exception
        with pytest.raises(ComputationError, match=msg.format(expr, chars)):
            rep.get(x_key)

    # Now using parseable but unrecognized units
    x["unit"] = "USD/kWa"
    scen.add_par("x", x)

    # Unrecognized units are added automatically, with log messages emitted
    caplog.clear()

    # Protect from --verbose command-line option, which sets the level to DEBUG
    with caplog.at_level(logging.INFO):
        rep.get(x_key)

    # NB cannot use assert_logs here. reporting.utils.parse_units uses the pint
    #    application registry, so depending which tests are run and in which order, this
    #    unit may already be defined.
    if len(caplog.messages):
        assert "Add unit definition: kWa = [kWa]" in caplog.messages

    # Mix of recognized/unrecognized units can be added: USD is already in the
    # unit registry, so is not re-added
    x["unit"] = "USD/pkm"
    test_mp.add_unit("USD/pkm")
    scen.add_par("x", x)

    caplog.clear()
    rep.get(x_key)
    assert not any("Add unit definition: USD = [USD]" in m for m in caplog.messages)

    # Mixed units are discarded
    x.loc[0, "unit"] = "kg"
    scen.add_par("x", x)

    with assert_logs(
        caplog, "x: mixed units ['kg', 'USD/pkm'] discarded", at_level=logging.INFO
    ):
        rep.get(x_key)

    # Configured unit substitutions are applied
    rep.graph["config"]["units"] = dict(apply=dict(x="USD/pkm"))

    with assert_logs(
        caplog, "x: replace units dimensionless with USD/pkm", at_level=logging.INFO
    ):
        x = rep.get(x_key)

    # Applied units are pint objects with the correct dimensionality
    unit = x.attrs["_unit"]
    assert isinstance(unit, pint.Unit)
    assert unit.dimensionality == {"[USD]": 1, "[pkm]": -1}


def test_cli(ixmp_cli, test_mp, test_data_path):
    # Put something in the database
    test_mp.open_db()
    make_dantzig(test_mp)
    test_mp.close_db()

    platform_name = test_mp.name

    # Delete the platform/close the database connection
    del test_mp

    cmd = [
        "--platform",
        platform_name,
        "--model",
        "canning problem",
        "--scenario",
        "standard",
        "report",
        "--config",
        str(test_data_path / "report-config-0.yaml"),
        "d_check",
    ]

    # 'report' command runs
    result = ixmp_cli.invoke(cmd)
    assert result.exit_code == 0

    # TODO warning should be logged

    # Reporting produces the expected command-line output

    assert re.match(
        "i          j       "  # Trailing whitespace
        r"""
san-diego  chicago     1\.8
           new-york    2\.5
           topeka      1\.4
seattle    chicago     1\.7
           new-york    2\.5
           topeka      1\.8
(Name: value, )?dtype: float64
""",
        result.output,
    )


def test_filters(test_mp, tmp_path, caplog):
    """Reporting can be filtered ex ante."""
    scen = ixmp.Scenario(test_mp, "Reporting filters", "Reporting filters", "new")
    t, t_foo, t_bar, x = add_test_data(scen)

    rep = Reporter.from_scenario(scen)
    x_key = rep.full_key("x")

    def assert_t_indices(labels):
        assert set(rep.get(x_key).coords["t"].values) == set(labels)

    # 1. Set filters directly
    rep.graph["config"]["filters"] = {"t": t_foo}
    assert_t_indices(t_foo)

    # Reporter can be re-used by changing filters
    rep.graph["config"]["filters"] = {"t": t_bar}
    assert_t_indices(t_bar)

    rep.graph["config"]["filters"] = {}
    assert_t_indices(t)

    # 2. Set filters using a convenience method
    rep = Reporter.from_scenario(scen)
    rep.set_filters(t=t_foo)
    assert_t_indices(t_foo)

    # Clear filters using the convenience method
    rep.set_filters(t=None)
    assert_t_indices(t)

    # Clear using the convenience method with no args
    rep.set_filters(t=t_foo)
    assert_t_indices(t_foo)
    rep.set_filters()
    assert_t_indices(t)

    # 3. Set filters via configuration keys
    # NB passes through from_scenario() -> __init__() -> configure()
    rep = Reporter.from_scenario(scen, filters={"t": t_foo})
    assert_t_indices(t_foo)

    # Configuration key can also be read from file
    rep = Reporter.from_scenario(scen)

    # Write a temporary file containing the desired labels
    config_file = tmp_path / "config.yaml"
    config_file.write_text("\n".join(["filters:", f"  t: {repr(t_bar)}"]))

    rep.configure(config_file)
    assert_t_indices(t_bar)

    # Filtering too heavily:
    # Remove one value from the database at valid coordinates
    removed = {"t": t[:1], "y": list(x.coords["y"].values)[:1]}
    scen.remove_par("x", removed)

    # Set filters to retrieve only this coordinate
    rep.set_filters(**removed)

    # A warning is logged
    with assert_logs(
        caplog,
        (
            f"0 values for par 'x' using filters: {repr(removed)}",
            "May be the cause of subsequent errors",
        ),
        at_level=logging.DEBUG,
    ):
        rep.get(x_key)



import gc
import logging
import os
import re
from collections import ChainMap
from collections.abc import Iterable, Sequence
from copy import copy
from itertools import chain
from pathlib import Path, PurePosixPath
from types import SimpleNamespace
from typing import Generator, Mapping
from weakref import WeakKeyDictionary

import jpype
import pandas as pd

from ixmp.core import Scenario
from ixmp.utils import as_str_list, filtered

from . import FIELDS, ItemType
from .base import CachingBackend

log = logging.getLogger(__name__)


_EXCEPTION_VERBOSE = os.environ.get("IXMP_JDBC_EXCEPTION_VERBOSE", "0") == "1"

# Whether to collect garbage aggressively when instances of TimeSeries die.
# See JDBCBackend.gc().
_GC_AGGRESSIVE = True

# Map of Python to Java log levels
# https://logging.apache.org/log4j/2.x/log4j-api/apidocs/org/apache/logging/log4j/Level.html
LOG_LEVELS = {
    "CRITICAL": "FATAL",
    "ERROR": "ERROR",
    "WARNING": "WARN",
    "INFO": "INFO",
    "DEBUG": "DEBUG",
    "NOTSET": "ALL",
}

# Java classes, loaded by start_jvm(). These become available as e.g.
# java.IxException or java.HashMap.
java = SimpleNamespace()

JAVA_CLASSES = [
    "at.ac.iiasa.ixmp.exceptions.IxException",
    "at.ac.iiasa.ixmp.objects.Scenario",
    "at.ac.iiasa.ixmp.dto.TimesliceDTO",
    "at.ac.iiasa.ixmp.Platform",
    "java.lang.Double",
    "java.lang.Exception",
    "java.lang.Integer",
    "java.lang.NoClassDefFoundError",
    "java.lang.IllegalArgumentException",
    "java.lang.Long",
    "java.lang.Runtime",
    "java.lang.System",
    "java.math.BigDecimal",
    "java.util.HashMap",
    "java.util.LinkedHashMap",
    "java.util.LinkedList",
    "java.util.ArrayList",
    "java.util.Properties",
    "at.ac.iiasa.ixmp.dto.DocumentationKey",
]


DRIVER_CLASS = {
    "oracle": "oracle.jdbc.driver.OracleDriver",
    "hsqldb": "org.hsqldb.jdbcDriver",
}


def _create_properties(driver=None, path=None, url=None, user=None, password=None):
    """Create a database Properties from arguments."""
    properties = java.Properties()

    # Handle arguments
    try:
        properties.setProperty("jdbc.driver", DRIVER_CLASS[driver])
    except KeyError:
        raise ValueError(f"unrecognized/unsupported JDBC driver {repr(driver)}")

    if driver == "oracle":
        if url is None or path is not None:
            raise ValueError("use JDBCBackend(driver='oracle', url=…)")

        full_url = f"jdbc:oracle:thin:@{url}"
    elif driver == "hsqldb":
        if path is None and url is None:
            raise ValueError("use JDBCBackend(driver='hsqldb', path=…)")

        if url is not None:
            if url.startswith("jdbc:hsqldb:"):
                full_url = url
            else:
                raise ValueError(url)
        else:
            # Convert Windows paths to use forward slashes per HyperSQL JDBC
            # URL spec
            url_path = str(PurePosixPath(Path(path).resolve())).replace("\\", "")
            full_url = f"jdbc:hsqldb:file:{url_path}"
        user = user or "ixmp"
        password = password or "ixmp"

    properties.setProperty("jdbc.url", full_url)
    properties.setProperty("jdbc.user", user)
    properties.setProperty("jdbc.pwd", password)

    return properties


def _read_properties(file):
    """Read database properties from *file*, returning :class:`dict`."""
    properties = dict()
    for line in file.read_text().split("\n"):
        match = re.search(r"([^\s]+)\s*=\s*(.+)\s*", line)
        if match is not None:
            properties[match.group(1)] = match.group(2)
    return properties


def _raise_jexception(exc, msg="unhandled Java exception: "):
    """Convert Java/JPype exceptions to ordinary Python RuntimeError."""
    # Try to re-raise as a ValueError for bad model or scenario name
    arg = exc.args[0] if isinstance(exc.args[0], str) else ""
    match = re.search(r"getting '([^']*)' in table '([^']*)'", arg)
    if match:
        param = match.group(2).lower()
        if param in {"model", "scenario"}:
            raise ValueError(f"{param}={repr(match.group(1))}") from None

    # Other exceptions
    if _EXCEPTION_VERBOSE:
        msg += "\n\n" + exc.stacktrace()
    else:
        msg += exc.message()

    raise RuntimeError(msg) from None


def _domain_enum(domain):
    domain_enum = java.DocumentationKey.DocumentationDomain
    try:
        return domain_enum.valueOf(domain.upper())
    except java.IllegalArgumentException:
        domains = ", ".join([d.name().lower() for d in domain_enum.values()])
        raise ValueError(f"No such domain: {domain}, " f"existing domains: {domains}")


def _unwrap(v):
    """Unwrap meta numeric value or list of values (BigDecimal -> Double)."""
    if isinstance(v, java.BigDecimal):
        return v.doubleValue()
    if isinstance(v, java.ArrayList):
        return [_unwrap(elt) for elt in v]
    return v


def _wrap(value):
    if isinstance(value, (str, bool)):
        return value
    if isinstance(value, (int, float)):
        return java.BigDecimal(value)
    elif isinstance(value, (Sequence, Iterable)):
        jlist = java.ArrayList()
        jlist.addAll([_wrap(elt) for elt in value])
        return jlist
    else:
        raise ValueError(f"Cannot use value {value} as metadata")


class JDBCBackend(CachingBackend):
    """Backend using JPype/JDBC to connect to Oracle and HyperSQL databases.

    Parameters
    ----------
    driver : 'oracle' or 'hsqldb'
        JDBC driver to use.
    path : path-like, optional
        Path to the HyperSQL database.
    url : str, optional
        Partial or complete JDBC URL for the Oracle or HyperSQL database,
        e.g. ``database-server.example.com:PORT:SCHEMA``. See
        :ref:`configuration`.
    user : str, optional
        Database user name.
    password : str, optional
        Database user password.
    cache : bool, optional
        If :obj:`True` (the default), cache Python objects after conversion
        from Java objects.
    jvmargs : str, optional
        Java Virtual Machine arguments. See :meth:`.start_jvm`.
    dbprops : path-like, optional
        With ``driver='oracle'``, the path to a database properties file
        containing `driver`, `url`, `user`, and `password` information.
    """

    # NB Much of the code of this backend is in Java, in the iiasa/ixmp_source
    #    Github repository.
    #
    #    Among other abstractions, this backend:
    #
    #    - Handles any conversion between Java and Python types that is not
    #      done automatically by JPype.
    #    - Catches Java exceptions such as ixmp.exceptions.IxException, and
    #      re-raises them as appropriate Python exceptions.
    #
    #    Limitations:
    #
    #    - s_clone() is only supported when target_backend is JDBCBackend.

    #: Reference to the at.ac.iiasa.ixmp.Platform Java object
    jobj: jpype.JObject = None

    #: Mapping from ixmp.TimeSeries object to the underlying
    #: at.ac.iiasa.ixmp.Scenario object (or subclasses of either)
    jindex: Mapping[object, jpype.JObject] = WeakKeyDictionary()

    def __init__(self, jvmargs=None, **kwargs):
        properties = None

        # Handle arguments
        if "dbprops" in kwargs:
            # Use an existing file
            dbprops = Path(kwargs.pop("dbprops"))
            if dbprops.exists() and dbprops.is_file():
                # Existing properties file
                properties = _read_properties(dbprops)
                if "jdbc.url" not in properties:
                    raise ValueError("Config file contains no database URL")
            else:
                raise FileNotFoundError(dbprops)

        start_jvm(jvmargs)

        # Invoke the parent constructor to initialize the cache
        super().__init__(cache_enabled=kwargs.pop("cache", True))

        # Extract a log_level keyword argument before _create_properties(). By default,
        # use the same level as the 'ixmp' logger, whatever that has been set to.
        ixmp_logger = logging.getLogger("ixmp")
        log_level = kwargs.pop("log_level", ixmp_logger.getEffectiveLevel())

        # Create a database properties object
        if properties:
            # ...using file contents
            new_props = java.Properties()
            [new_props.setProperty(k, v) for k, v in properties.items()]
            properties = new_props
        else:
            # ...from arguments
            try:
                properties = _create_properties(**kwargs)
            except TypeError as e:
                msg = e.args[0].replace("_create_properties", "JDBCBackend")
                raise TypeError(msg)

        log.info(
            "launching ixmp.Platform connected to {}".format(
                properties.getProperty("jdbc.url")
            )
        )

        try:
            self.jobj = java.Platform("Python", properties)
        except java.NoClassDefFoundError as e:  # pragma: no cover
            raise NameError(
                f"{e}\nCheck that dependencies of ixmp.jar are "
                f"included in {Path(__file__).parents[2] / 'lib'}"
            )
        except jpype.JException as e:  # pragma: no cover
            # Handle Java exceptions
            jclass = e.__class__.__name__
            if jclass.endswith("HikariPool.PoolInitializationException"):
                redacted = copy(kwargs)
                redacted.update({"user": "(HIDDEN)", "password": "(HIDDEN)"})
                msg = f"unable to connect to database:\n{repr(redacted)}"
            elif jclass.endswith("FlywayException"):
                msg = "when initializing database:"
                if "applied migration" in e.args[0]:
                    msg += (
                        "\n\nThe schema of the database does not match the "
                        "schema of this version of ixmp. To resolve, either "
                        "install the version of ixmp used to create the "
                        "database, or delete it and retry."
                    )
            else:
                _raise_jexception(e)
            raise RuntimeError(f"{msg}\n(Java: {jclass})")

        # Set the log level
        self.set_log_level(log_level)

    def __del__(self):
        self.close_db()

    @classmethod
    def gc(cls):
        if _GC_AGGRESSIVE:
            # log.debug('Collect garbage')
            try:
                java.System.gc()
            except jpype.JVMNotRunning:
                pass
            gc.collect()
        # else:
        #     log.debug('Skip garbage collection')

    # Platform methods
    @classmethod
    def handle_config(cls, args, kwargs):
        """Handle platform/backend config arguments.

        `args` will overwrite any `kwargs`, and may be one of:

        - ("oracle", url, user, password) for an Oracle database.
        - ("hsqldb", path) for a file-backed HyperSQL database.
        - ("hsqldb",) with "url" supplied via `kwargs`, e.g. "jdbc:hsqldb:mem://foo" for
          an in-memory database.
        """
        args = list(args)
        info = copy(kwargs)

        # First argument: driver
        try:
            info["driver"] = args.pop(0)
        except IndexError:
            raise ValueError(
                f"≥1 positional argument required for class=jdbc: driver; got: {args}, "
                + str(kwargs)
            )

        if info["driver"] == "oracle":
            if len(args) != 3:
                raise ValueError(
                    "3 arguments required for driver=oracle: url, user, password; got: "
                    + str(args)
                )
            info["url"], info["user"], info["password"] = args

        elif info["driver"] == "hsqldb":
            try:
                info["path"] = Path(args.pop(0)).resolve()
            except IndexError:
                if "url" not in info:
                    raise ValueError(
                        "must supply either positional path or url= keyword argument "
                        "for driver=hsqldb"
                    )

            if len(args):
                raise ValueError(
                    f"Unrecognized extra argument(s) for driver=hsqldb: {args}"
                )
        else:
            raise ValueError(
                f"driver={info['driver']}; expected one of {set(DRIVER_CLASS)}"
            )

        return info

    def set_log_level(self, level):
        # Set the level of the 'ixmp.backend.jdbc' logger. Messages are handled by the
        # 'ixmp' logger; see ixmp/__init__.py.
        log.setLevel(level)

        # Translate to Java log level and set
        if isinstance(level, int):
            level = logging.getLevelName(level)
        self.jobj.setLogLevel(LOG_LEVELS[level])

    def get_log_level(self):
        levels = {v: k for k, v in LOG_LEVELS.items()}
        return levels.get(self.jobj.getLogLevel(), "UNKNOWN")

    def set_doc(self, domain, docs):
        dd = _domain_enum(domain)
        jdata = java.LinkedHashMap()
        if type(docs) == dict:
            docs = list(docs.items())
        for k, v in docs:
            jdata.put(str(k), str(v))
        self.jobj.setDoc(dd, jdata)

    def get_doc(self, domain, name=None):
        dd = _domain_enum(domain)
        if name is None:
            doc = self.jobj.getDoc(dd)
            return {entry.getKey(): entry.getValue() for entry in doc.entrySet()}
        else:
            return self.jobj.getDoc(dd, str(name))

    def open_db(self):
        """(Re-)open the database connection."""
        self.jobj.openDB()

    def close_db(self):
        """Close the database connection.

        A HSQL database can only be used by one :class:`Backend` instance at a
        time. Any existing connection must be closed before a new one can be
        opened.
        """
        try:
            self.jobj.closeDB()
        except java.IxException as e:  # pragma: no cover
            log.warning(str(e))
        except (AttributeError, jpype.JVMNotRunning):
            # - self.jobj is None, e.g. cleanup after __init__ fails
            # - JVM has already shut down, e.g. on program exit
            pass

    def get_auth(self, user, models, kind):
        return self.jobj.checkModelAccess(user, kind, to_jlist(models))

    def set_node(self, name, parent=None, hierarchy=None, synonym=None):
        if parent and hierarchy and not synonym:
            self.jobj.addNode(name, parent, hierarchy)
        elif synonym and not (parent or hierarchy):
            self.jobj.addNodeSynonym(synonym, name)

    def get_nodes(self):
        for r in self.jobj.listNodes("%"):
            n, p, h = r.getName(), r.getParent(), r.getHierarchy()
            yield (n, None, p, h)
            yield from [(s, n, p, h) for s in (r.getSynonyms() or [])]

    def get_timeslices(self):
        for r in self.jobj.getTimeslices():
            name, category, duration = (r.getName(), r.getCategory(), r.getDuration())
            yield name, category, duration

    def set_timeslice(self, name, category, duration):
        self.jobj.addTimeslice(name, category, java.Double(duration))

    def add_model_name(self, name):
        self.jobj.addModel(str(name))

    def add_scenario_name(self, name):
        self.jobj.addScenario(str(name))

    def get_model_names(self) -> Generator[str, None, None]:
        for model in self.jobj.listModels():
            yield str(model)

    def get_scenario_names(self) -> Generator[str, None, None]:
        for scenario in self.jobj.listScenarios():
            yield str(scenario)

    def get_scenarios(self, default, model, scenario):
        # List<Map<String, Object>>
        try:
            scenarios = self.jobj.getScenarioList(default, model, scenario)
        except java.IxException as e:
            _raise_jexception(e)

        for s in scenarios:
            data = []
            for field in FIELDS["get_scenarios"]:
                data.append(int(s[field]) if field == "version" else s[field])
            yield data

    def set_unit(self, name, comment):
        self.jobj.addUnitToDB(name, comment)

    def get_units(self):
        return to_pylist(self.jobj.getUnitList())

    def read_file(self, path, item_type: ItemType, **kwargs):
        """Read Platform, TimeSeries, or Scenario data from file.

        JDBCBackend supports reading from:

        - ``path='*.gdx', item_type=ItemType.MODEL``. The keyword arguments
          `check_solution`, `comment`, `equ_list`, and `var_list` are
          **required**.

        Other parameters
        ----------------
        check_solution : bool
            If True, raise an exception if the GAMS solver did not reach
            optimality. (Only for MESSAGE-scheme Scenarios.)
        comment : str
            Comment added to Scenario when importing the solution.
        equ_list : list of str
            Equations to be imported.
        var_list : list of str
            Variables to be imported.
        filters : dict of dict of str
            Restrict items read.

        See also
        --------
        .Backend.read_file
        """
        try:
            # Call the default implementation, e.g. for .xlsx
            super().read_file(path, item_type, **kwargs)
        except NotImplementedError:
            pass
        else:
            return

        ts, filters = self._handle_rw_filters(kwargs.pop("filters", {}))
        if path.suffix == ".gdx" and item_type is ItemType.MODEL:
            kw = {"check_solution", "comment", "equ_list", "var_list"}

            if not isinstance(ts, Scenario):
                raise ValueError("read from GDX requires a Scenario object")
            elif set(kwargs.keys()) != kw:
                raise ValueError(
                    f"keyword arguments {kwargs.keys()} do not " f"match required {kw}"
                )

            args = (
                str(path.parent),
                path.name,
                kwargs.pop("comment"),
                to_jlist(kwargs.pop("var_list")),
                to_jlist(kwargs.pop("equ_list")),
                kwargs.pop("check_solution"),
            )

            if len(kwargs):
                raise ValueError(f"extra keyword arguments {kwargs}")

            self.jindex[ts].readSolutionFromGDX(*args)

            self.cache_invalidate(ts)
        else:
            raise NotImplementedError(path, item_type)

    def write_file(self, path, item_type: ItemType, **kwargs):
        """Write Platform, TimeSeries, or Scenario data to file.

        JDBCBackend supports writing to:

        - ``path='*.gdx', item_type=ItemType.SET | ItemType.PAR``.
        - ``path='*.csv', item_type=TS``. The `default` keyword argument is
          **required**.

        Other parameters
        ----------------
        filters : dict of dict of str
            Restrict items written. The following filters may be used:

            - model : str
            - scenario : str
            - variable : list of str
            - default : bool. If :obj:`True`, only data from TimeSeries
              versions with :meth:`set_default` are written.

        See also
        --------
        .Backend.write_file
        """
        try:
            # Call the default implementation, e.g. for .xlsx
            super().write_file(path, item_type, **kwargs)
        except NotImplementedError:
            pass
        else:
            return

        ts, filters = self._handle_rw_filters(kwargs.pop("filters", {}))
        if path.suffix == ".gdx" and item_type is ItemType.SET | ItemType.PAR:
            if len(filters):
                raise NotImplementedError("write to GDX with filters")
            elif not isinstance(ts, Scenario):
                raise ValueError("write to GDX requires a Scenario object")

            # include_var_equ=False -> do not include variables/equations in
            # GDX
            self.jindex[ts].toGDX(str(path.parent), path.name, False)
        elif path.suffix == ".csv" and item_type is ItemType.TS:
            models = set(filters.pop("model"))
            scenarios = set(filters.pop("scenario"))
            variables = filters.pop("variable")
            units = filters.pop("unit")
            regions = filters.pop("region")
            default = filters.pop("default")
            export_all_runs = filters.pop("export_all_runs")

            scen_list = self.jobj.getScenarioList(default, None, None)
            # TODO: replace with passing list of models/scenarios
            #       to the method above
            run_ids = [
                s["run_id"]
                for s in scen_list
                if (len(scenarios) == 0 or s["scenario"] in scenarios)
                and (len(models) == 0 or s["model"] in models)
            ]
            self.jobj.exportTimeseriesData(
                to_jlist(run_ids),
                to_jlist(variables),
                to_jlist(units),
                to_jlist(regions),
                str(path),
                export_all_runs,
            )
        else:
            raise NotImplementedError

    # Timeseries methods

    def _index_and_set_attrs(self, jobj, ts):
        """Add *jobj* to index and update attributes of *ts*.

        Helper for init and get.
        """
        # Add to index
        self.jindex[ts] = jobj

        # Retrieve the version of the Java object
        v = jobj.getVersion()
        if ts.version is None:
            # The default version was requested; update the attribute
            ts.version = v
        elif v != ts.version:  # pragma: no cover
            # Something went wrong on the Java side
            raise RuntimeError(f"got version {v} instead of {ts.version}")

        if isinstance(ts, Scenario):
            # Also retrieve the scheme
            s = jobj.getScheme()

            if ts.scheme and s != ts.scheme:  # pragma: no cover
                # Something went wrong on the Java side
                raise RuntimeError(f"got scheme {s} instead of {ts.scheme}")

            ts.scheme = s

    def _validate_meta_args(self, model, scenario, version):
        """Validate arguments for getting/setting/deleting meta"""
        valid = False
        if model and not scenario and version is None:
            valid = True
        elif scenario and not model and version is None:
            valid = True
        elif model and scenario and version is None:
            valid = True
        elif model and scenario and version is not None:
            valid = True
        if not valid:
            msg = (
                "Invalid arguments. Valid combinations are: (model), "
                "(scenario), (model, scenario), (model, scenario, version)"
            )
            raise ValueError(msg)

    def init(self, ts, annotation):
        klass = ts.__class__.__name__

        # Final arguments: scheme only for Scenarios
        args = [ts.scheme, annotation] if klass == "Scenario" else [annotation]

        # Call either newTimeSeries or newScenario
        method = getattr(self.jobj, "new" + klass)
        jobj = method(ts.model, ts.scenario, *args)

        self._index_and_set_attrs(jobj, ts)

    def get(self, ts):
        args = [ts.model, ts.scenario]
        if ts.version is not None:
            # Load a TimeSeries of specific version
            args.append(ts.version)

        # either getTimeSeries or getScenario
        method = getattr(self.jobj, "get" + ts.__class__.__name__)
        try:
            # Either the 2- or 3- argument form, depending on args
            jobj = method(*args)
        except java.IxException as e:
            # Re-raise as a ValueError for bad model or scenario name, or other
            _raise_jexception(e)

        self._index_and_set_attrs(jobj, ts)

    def del_ts(self, ts):
        super().del_ts(ts)

        # Aggressively free memory
        self.gc()

    def check_out(self, ts, timeseries_only):
        try:
            self.jindex[ts].checkOut(timeseries_only)
        except java.IxException as e:
            _raise_jexception(e)

    def commit(self, ts, comment):
        try:
            self.jindex[ts].commit(comment)
        except java.Exception as e:
            arg = e.args[0]
            if isinstance(arg, str) and "this Scenario is not checked out" in arg:
                raise RuntimeError(arg)
            else:  # pragma: no cover
                _raise_jexception(e)
        if ts.version == 0:
            ts.version = self.jindex[ts].getVersion()

    def discard_changes(self, ts):
        self.jindex[ts].discardChanges()

    def set_as_default(self, ts):
        self.jindex[ts].setAsDefaultVersion()

    def is_default(self, ts):
        return bool(self.jindex[ts].isDefault())

    def last_update(self, ts):
        timestamp = self.jindex[ts].getLastUpdateTimestamp()
        if timestamp is not None:
            return timestamp.toString()
        else:
            return timestamp  # None

    def run_id(self, ts):
        return self.jindex[ts].getRunId()

    def preload(self, ts):
        self.jindex[ts].preloadAllTimeseries()

    def get_data(self, ts, region, variable, unit, year):
        # Convert the selectors to Java lists
        r = to_jlist(region)
        v = to_jlist(variable)
        u = to_jlist(unit)
        y = to_jlist(year)

        # Field types
        ftype = {
            "year": int,
            "value": float,
        }

        # Iterate over returned rows
        for row in self.jindex[ts].getTimeseries(r, v, u, None, y):
            # Get the value of each field and maybe convert its type
            yield tuple(
                ftype.get(f, str)(getattr(row, "get" + f.capitalize())())
                for f in FIELDS["ts_get"]
            )

    def get_geo(self, ts):
        # NB the return type of getGeoData() requires more processing than
        #    getTimeseries. It also accepts no selectors.

        # Field types
        ftype = {
            "meta": int,
            "year": lambda obj: obj,  # Pass through; handled later
        }

        # Returned names in Java data structure do not match API column names
        jname = {
            "meta": "meta",
            "region": "nodeName",
            "subannual": "subannual",
            "unit": "unitName",
            "variable": "keyString",
            "year": "yearlyData",
        }

        # Iterate over rows from the Java backend
        for row in self.jindex[ts].getGeoData():
            data1 = {
                f: ftype.get(f, str)(row.get(jname.get(f, f)))
                for f in FIELDS["ts_get_geo"]
                if f != "value"
            }

            # At this point, the 'year' key is a not a single value, but a
            # year -> value mapping with multiple entries
            yv_entries = data1.pop("year").entrySet()

            # Construct a chain map: look up in data1, then data2
            data2 = {"year": None, "value": None}
            cm = ChainMap(data1, data2)

            for yv in yv_entries:
                # Update data2
                data2["year"] = yv.getKey()
                data2["value"] = yv.getValue()

                # Construct a row with a single value
                yield tuple(cm[f] for f in FIELDS["ts_get_geo"])

    def set_data(self, ts, region, variable, data, unit, subannual, meta):
        # Convert *data* to a Java data structure
        jdata = java.LinkedHashMap()
        for k, v in data.items():
            # Explicit cast is necessary; otherwise java.lang.Long
            jdata.put(java.Integer(k), v)

        self.jindex[ts].addTimeseries(region, variable, subannual, jdata, unit, meta)

    def set_geo(self, ts, region, variable, subannual, year, value, unit, meta):
        self.jindex[ts].addGeoData(
            region, variable, subannual, java.Integer(year), value, unit, meta
        )

    def delete(self, ts, region, variable, subannual, years, unit):
        years = to_jlist(years, java.Integer)
        self.jindex[ts].removeTimeseries(region, variable, subannual, years, unit)

    def delete_geo(self, ts, region, variable, subannual, years, unit):
        years = to_jlist(years, java.Integer)
        self.jindex[ts].removeGeoData(region, variable, subannual, years, unit)

    # Scenario methods

    def clone(
        self,
        s,
        platform_dest,
        model,
        scenario,
        annotation,
        keep_solution,
        first_model_year=None,
    ):
        # Raise exceptions for limitations of JDBCBackend
        if not isinstance(platform_dest._backend, self.__class__):
            raise NotImplementedError(  # pragma: no cover
                f"Clone between {self.__class__} and"
                f"{platform_dest._backend.__class__}"
            )
        elif platform_dest._backend is not self:
            package = s.__class__.__module__.split(".")[0]
            msg = f"Cross-platform clone of {package}.Scenario with"
            if keep_solution is False:
                raise NotImplementedError(f"{msg} `keep_solution=False`")
            elif "message_ix" in msg and first_model_year is not None:
                raise NotImplementedError(f"{msg} first_model_year != None")

        # Prepare arguments
        args = [platform_dest._backend.jobj, model, scenario, annotation, keep_solution]
        if first_model_year:
            args.append(first_model_year)

        # Reference to the cloned Java object
        jclone = self.jindex[s].clone(*args)

        # Instantiate same class as the original object
        return s.__class__(
            platform_dest,
            model,
            scenario,
            version=jclone.getVersion(),
            scheme=jclone.getScheme(),
        )

    def has_solution(self, s):
        return self.jindex[s].hasSolution()

    def list_items(self, s, type):
        return to_pylist(getattr(self.jindex[s], f"get{type.title()}List")())

    def init_item(self, s, type, name, idx_sets, idx_names):
        # generate index-set and index-name lists
        if isinstance(idx_sets, set) or isinstance(idx_names, set):
            raise TypeError("index dimension must be string or ordered lists")

        idx_sets = to_jlist(idx_sets) if len(idx_sets) else None

        if idx_names:
            if len(idx_names) != len(idx_sets):
                raise ValueError(
                    f"index names {repr(idx_names)} must have same length as "
                    f"index sets {repr(idx_sets)}"
                )
            idx_names = to_jlist(idx_names)
        else:
            idx_names = idx_sets

        # Initialize the Item
        func = getattr(self.jindex[s], f"initialize{type.title()}")

        # The constructor returns a reference to the Java Item, but these
        # aren't exposed by Backend, so don't return here
        try:
            func(name, idx_sets, idx_names)
        except jpype.JException as e:
            if "already exists" in e.args[0]:
                raise ValueError(f"{repr(name)} already exists")
            else:
                _raise_jexception(e)

    def delete_item(self, s, type, name):
        getattr(self.jindex[s], f"remove{type.title()}")(name)
        self.cache_invalidate(s, type, name)

    def item_index(self, s, name, sets_or_names):
        jitem = self._get_item(s, "item", name, load=False)
        return list(getattr(jitem, f"getIdx{sets_or_names.title()}")())

    def item_get_elements(self, s, type, name, filters=None):
        if filters:
            # Convert filter elements to strings
            filters = {dim: as_str_list(ele) for dim, ele in filters.items()}

        try:
            # Retrieve the cached value with this exact set of filters
            return self.cache_get(s, type, name, filters)
        except KeyError:
            pass  # Cache miss

        try:
            # Retrieve a cached, unfiltered value of the same item
            unfiltered = self.cache_get(s, type, name, None)
        except KeyError:
            pass  # Cache miss
        else:
            # Success; filter and return
            return filtered(unfiltered, filters)

        # Failed to load item from cache

        # Retrieve the item
        item = self._get_item(s, type, name, load=True)
        idx_names = list(item.getIdxNames())
        idx_sets = list(item.getIdxSets())

        # Get list of elements, using filters if provided
        if filters is not None:
            jFilter = java.HashMap()

            for idx_name, values in filters.items():
                # Retrieve the elements of the index set as a list
                idx_set = idx_sets[idx_names.index(idx_name)]
                elements = self.item_get_elements(s, "set", idx_set).tolist()

                # Filter for only included values and store
                filtered_elements = filter(lambda e: e in values, elements)
                jFilter.put(idx_name, to_jlist(filtered_elements))

            jList = item.getElements(jFilter)
        else:
            jList = item.getElements()

        if item.getDim() > 0:
            # Mapping set or multi-dimensional equation, parameter, or variable
            columns = copy(idx_names)

            # Prepare dtypes for index columns
            dtypes = {}
            for idx_name, idx_set in zip(columns, idx_sets):
                # NB using categoricals could be more memory-efficient, but
                #    requires adjustment of tests/documentation. See
                #    https://github.com/iiasa/ixmp/issues/228
                # dtypes[idx_name] = CategoricalDtype(
                #     self.item_get_elements(s, 'set', idx_set))
                dtypes[idx_name] = str

            # Prepare dtypes for additional columns
            if type == "par":
                columns.extend(["value", "unit"])
                dtypes["value"] = float
                # Same as above
                # dtypes['unit'] = CategoricalDtype(self.jobj.getUnitList())
                dtypes["unit"] = str
            elif type in ("equ", "var"):
                columns.extend(["lvl", "mrg"])
                dtypes.update({"lvl": float, "mrg": float})
            # Prepare empty DataFrame
            result = pd.DataFrame(index=pd.RangeIndex(len(jList)), columns=columns)

            # Copy vectors from Java into DataFrame columns
            # NB [:] causes JPype to use a faster code path
            for i in range(len(idx_sets)):
                result.iloc[:, i] = item.getCol(i, jList)[:]
            if type == "par":
                result.loc[:, "value"] = item.getValues(jList)[:]
                result.loc[:, "unit"] = item.getUnits(jList)[:]
            elif type in ("equ", "var"):
                result.loc[:, "lvl"] = item.getLevels(jList)[:]
                result.loc[:, "mrg"] = item.getMarginals(jList)[:]

            # .loc assignment above modifies dtypes; set afterwards
            result = result.astype(dtypes)
        elif type == "set":
            # Index sets
            # dtype=object is to silence a warning in pandas 1.0
            result = pd.Series(item.getCol(0, jList), dtype=object)
        elif type == "par":
            # Scalar parameters
            result = dict(
                value=float(item.getScalarValue().floatValue()),
                unit=str(item.getScalarUnit()),
            )
        elif type in ("equ", "var"):
            # Scalar equations and variables
            result = dict(
                lvl=float(item.getScalarLevel().floatValue()),
                mrg=float(item.getScalarMarginal().floatValue()),
            )

        # Store cache
        self.cache(s, type, name, filters, result)

        return result

    def item_set_elements(self, s, type, name, elements):
        jobj = self._get_item(s, type, name)

        try:
            for key, value, unit, comment in elements:
                # Prepare arguments
                args = [to_jlist(key)] if key else []
                if type == "par":
                    args.extend([java.Double(value), unit])
                if comment:
                    args.append(comment)

                # Activates one of 5 signatures for addElement:
                # - set: (key)
                # - set: (key, comment)
                # - par: (key, value, unit, comment)
                # - par: (key, value, unit)
                # - par: (value, unit, comment)
                jobj.addElement(*args)
        except java.IxException as e:
            if any(s in e.args[0] for s in ("does not have an element", "The unit")):
                # Re-raise as Python ValueError
                raise ValueError(e.args[0]) from None
            elif "cannot be edited" in e.args[0]:
                raise RuntimeError(e.args[0])
            else:  # pragma: no cover
                _raise_jexception(e)

        self.cache_invalidate(s, type, name)

    def item_delete_elements(self, s, type, name, keys):
        jitem = self._get_item(s, type, name, load=False)
        for key in keys:
            jitem.removeElement(to_jlist(key))

        # Since `name` may be an index set, clear the cache entirely. This
        # ensures that e.g. parameter elements for parameters indexed by `name`
        # are also refreshed on the next call to item_get_elements.
        args = (s,) if type == "set" else (s, type, name)
        self.cache_invalidate(*args)

    def get_meta(
        self,
        model: str = None,
        scenario: str = None,
        version: int = None,
        strict: bool = False,
    ) -> dict:
        self._validate_meta_args(model, scenario, version)
        if version is not None:
            version = java.Long(version)

        try:
            meta = self.jobj.getMeta(model, scenario, version, strict)
        except java.IxException as e:
            _raise_jexception(e)

        return {entry.getKey(): _unwrap(entry.getValue()) for entry in meta.entrySet()}

    def set_meta(
        self, meta: dict, model: str = None, scenario: str = None, version: int = None
    ) -> None:
        self._validate_meta_args(model, scenario, version)
        if version is not None:
            version = java.Long(version)

        jmeta = java.HashMap()
        for k, v in meta.items():
            jmeta.put(str(k), _wrap(v))

        try:
            self.jobj.setMeta(model, scenario, version, jmeta)
        except java.IxException as e:
            _raise_jexception(e)

    def remove_meta(
        self, categories, model: str = None, scenario: str = None, version: int = None
    ):
        self._validate_meta_args(model, scenario, version)
        if version is not None:
            version = java.Long(version)
        return self.jobj.removeMeta(model, scenario, version, to_jlist(categories))

    def clear_solution(self, s, from_year=None):
        from ixmp.core import Scenario

        if from_year:
            if type(s) is not Scenario:
                raise TypeError(
                    "s_clear_solution(from_year=...) only valid "
                    "for ixmp.Scenario; not subclasses"
                )
            self.jindex[s].removeSolution(from_year)
        else:
            self.jindex[s].removeSolution()

        self.cache_invalidate(s)

    # MsgScenario methods

    def cat_list(self, ms, name):
        return to_pylist(self.jindex[ms].getTypeList(name))

    def cat_get_elements(self, ms, name, cat):
        return to_pylist(self.jindex[ms].getCatEle(name, cat))

    def cat_set_elements(self, ms, name, cat, keys, is_unique):
        self.jindex[ms].addCatEle(name, cat, to_jlist(keys), is_unique)

    # Helpers; not part of the Backend interface

    def _get_item(self, s, ix_type, name, load=True):
        """Return the Java object for item *name* of *ix_type*.

        Parameters
        ----------
        load : bool, optional
            If *ix_type* is 'par', 'var', or 'equ', the elements of the item
            are loaded from the database before :meth:`_item` returns. If
            :const:`False`, the elements can be loaded later using
            ``item.loadItemElementsfromDB()``.
        """
        # getItem is not overloaded to accept a second bool argument
        args = [name] + ([load] if ix_type != "item" else [])
        try:
            return getattr(self.jindex[s], f"get{ix_type.title()}")(*args)
        except java.IxException as e:
            # Regex for similar but not consistent messages from Java code
            msg = f"No (item|{ix_type.title()}) '?{name}'? exists in this " "Scenario!"
            if re.match(msg, e.args[0]):
                # Re-raise as a Python KeyError
                raise KeyError(name) from None
            else:  # pragma: no cover
                _raise_jexception(e)


def start_jvm(jvmargs=None):
    """Start the Java Virtual Machine via :mod:`JPype`.

    Parameters
    ----------
    jvmargs : str or list of str, optional
        Additional arguments for launching the JVM, passed to
        :func:`jpype.startJVM`.

        For instance, to set the maximum heap space to 4 GiB, give
        ``jvmargs=['-Xmx4G']``. See the `JVM documentation`_ for a list of
        options.

        .. _`JVM documentation`: https://docs.oracle.com/javase/7/docs
           /technotes/tools/windows/java.html)
    """
    if jvmargs is None:
        jvmargs = []
    if jpype.isJVMStarted():
        return

    # Arguments
    args = jvmargs if isinstance(jvmargs, list) else [jvmargs]

    # Base for Java classpath entries
    cp = Path(__file__).parents[1]

    # Keyword arguments
    kwargs = dict(
        # Given 'lib/*' JPype will only glob '*.jar', so glob here explicitly
        classpath=map(str, chain([cp / "ixmp.jar"], cp.glob("lib/*"))),
        # For JPype 0.7 (raises a warning) and 0.8 (default is False).
        # 'True' causes Java string objects to be converted automatically to
        # Python str(), as expected by ixmp Python code.
        convertStrings=True,
    )

    log.debug(f"JAVA_HOME: {os.environ.get('JAVA_HOME', '(not set)')}")
    log.debug(f"jpype.getDefaultJVMPath: {jpype.getDefaultJVMPath()}")
    log.debug(f"args to startJVM: {args} {kwargs}")

    try:
        jpype.startJVM(*args, **kwargs)
    except FileNotFoundError as e:  # pragma: no cover
        # Not covered by tests. jpype.getDefaultJVMPath() tries an extensive set of
        # methods to find the JVM; it would require excessive effort to defeat these.
        raise FileNotFoundError(
            "This error may occur because you have not installed or configured a Java"
            "Runtime Environment. See the install documentation."
        ) from e

    # define auxiliary references to Java classes
    global java
    for class_name in JAVA_CLASSES:
        setattr(java, class_name.split(".")[-1], jpype.JClass(class_name))


# Conversion methods


def to_pylist(jlist):
    """Convert Java list types to :class:`list`."""
    try:
        return list(jlist[:])
    except Exception:
        # java.LinkedList
        return list(jlist.toArray()[:])


def to_jlist(arg, convert=None):
    """Convert :class:`list` *arg* to java.LinkedList.

    Parameters
    ----------
    arg : Collection or Iterable or str
    convert : callable, optional
        If supplied, every element of *arg* is passed through `convert` before
        being added.

    Returns
    -------
    java.LinkedList
    """
    jlist = java.LinkedList()

    # Previously JPype1 (prior to 1.0) could take single argument
    # in addAll method of Java collection. As string implements Sequence
    # contract in Python we need to convert it explicitly to list here.
    if isinstance(arg, str):
        arg = [arg]

    if convert:
        arg = map(convert, arg)

    if isinstance(arg, Sequence):
        # Sized collection can be used directly
        jlist.addAll(arg)

    elif isinstance(arg, Iterable):
        # Transfer items from an iterable, generator, etc. to the LinkedList
        [jlist.add(value) for value in arg]
    else:
        raise ValueError(arg)

    return jlist



import ixmp
from ixmp.testing import add_random_model_data, models


def test_read_excel_big(test_mp, tmp_path):
    """Excel files with model items split across sheets can be read.

    https://github.com/iiasa/ixmp/pull/345.
    """
    tmp_path /= "output.xlsx"

    # Write a 25-element parameter with max_row=10 → split across 3 sheets
    scen = ixmp.Scenario(test_mp, **models["dantzig"], version="new")
    add_random_model_data(scen, 25)
    scen.to_excel(tmp_path, items=ixmp.ItemType.MODEL, max_row=10)

    # Initialize target scenario for reading
    scen_empty = ixmp.Scenario(test_mp, "foo", "bar", version="new")
    scen_empty.init_set("random_set")
    scen_empty.init_par(
        "random_par", scen.idx_sets("random_par"), scen.idx_names("random_par")
    )

    # File can be read
    scen_empty.read_excel(tmp_path)

    assert len(scen_empty.par("random_par")) == 25



from typing import Dict, Type

#: Mapping from names to available models. To register additional models,
#: add elements to this variable.
MODELS: Dict[str, Type] = {}


def get_model(name, **model_options):
    """Return a model for *name* (or the default) with *model_options*."""
    try:
        return MODELS[name](**model_options)
    except KeyError:
        return MODELS["default"](name=name, **model_options)



import ixmp
from ixmp.testing import models
from ixmp.testing.data import random_ts_data
from ixmp.testing.resource import memory_usage


def test_random_ts_data(N=100):
    df = random_ts_data(N)
    assert N == len(df)


def test_resource_limit(resource_limit, test_mp):
    """Exercise :func:`memory_usage` and :func:`resource_limit`."""
    # TODO expand to cover other missed lines in those functions

    memory_usage("setup")

    s = ixmp.TimeSeries(test_mp, **models["h2g2"], version="new")

    memory_usage("1 TimeSeries")

    del s

    # commented: this is nondeterministic, not always true
    # assert info0.python <= info1.python
